from dataclasses import dataclass, field
from typing import Any, Optional, List

from apipeline.frames.control_frames import ControlFrame

from achatbot.common.types import VADAnalyzerArgs
from achatbot.types.speech.language import Language

#
# Control frames
#


@dataclass
class LLMFullResponseStartFrame(ControlFrame):
    """Used to indicate the beginning of a full LLM response. Following
    LLMResponseStartFrame, TextFrame and LLMResponseEndFrame for each sentence
    until a LLMFullResponseEndFrame."""

    pass


@dataclass
class LLMFullResponseEndFrame(ControlFrame):
    """Indicates the end of a full LLM response."""

    pass


@dataclass
class LLMResponseStartFrame(ControlFrame):
    """Used to indicate the beginning of an LLM response. Following TextFrames
    are part of the LLM response until an LLMResponseEndFrame"""

    pass


@dataclass
class LLMResponseEndFrame(ControlFrame):
    """Indicates the end of an LLM response."""

    pass


@dataclass
class UserStartedSpeakingFrame(ControlFrame):
    """Emitted by VAD to indicate that a user has started speaking. This can be
    used for interruptions or other times when detecting that someone is
    speaking is more important than knowing what they're saying (as you will
    with a TranscriptionFrame)

    """

    pass


@dataclass
class UserStoppedSpeakingFrame(ControlFrame):
    """Emitted by the VAD to indicate that a user stopped speaking."""

    pass


@dataclass
class TTSStartedFrame(ControlFrame):
    """Used to indicate the beginning of a TTS response. Following
    AudioRawFrames are part of the TTS response until an TTSEndFrame. These
    frames can be used for aggregating audio frames in a transport to optimize
    the size of frames sent to the session, without needing to control this in
    the TTS service.

    """

    pass


@dataclass
class TTSStoppedFrame(ControlFrame):
    """Indicates the end of a TTS response."""

    pass


@dataclass
class UserImageRequestFrame(ControlFrame):
    """A frame user to request an image from the given user."""

    user_id: str
    context: Optional[Any] = None  # for openai llm context

    def __str__(self):
        return f"{self.name}, user: {self.user_id}"


@dataclass
class BotStartedSpeakingFrame(ControlFrame):
    """Emitted upstream by transport outputs to indicate the bot started speaking."""

    pass


@dataclass
class BotStoppedSpeakingFrame(ControlFrame):
    """Emitted upstream by transport outputs to indicate the bot stopped speaking."""

    pass


@dataclass
class BotSpeakingFrame(ControlFrame):
    """Emitted by transport outputs while the bot is still speaking. This can be
    used, for example, to detect when a user is idle. That is, while the bot is
    speaking we don't want to trigger any user idle timeout since the user might
    be listening.

    """

    pass


@dataclass
class LLMModelUpdateFrame(ControlFrame):
    """A control frame containing a request to update to a new LLM model."""

    model: str


@dataclass
class TTSVoiceUpdateFrame(ControlFrame):
    """A control frame containing a request to update to a new TTS voice."""

    voice: str


@dataclass
class VADParamsUpdateFrame(ControlFrame):
    """A control frame containing a request to update VAD params. Intended
    to be pushed upstream from RTVI processor.
    """

    params: VADAnalyzerArgs


@dataclass
class ASRModelUpdateFrame(ControlFrame):
    """A control frame containing a request to update the ASR model and optional
    language.
    """

    model: str


@dataclass
class ASRLanguageUpdateFrame(ControlFrame):
    """A control frame containing a request to update to ASR language."""

    language: Language


@dataclass
class ASRArgsUpdateFrame(ControlFrame):
    """A control frame containing a request to update to ASR args."""

    args: dict


@dataclass
class TTSLanguageUpdateFrame(ControlFrame):
    """A control frame containing a request to update to a new TTS language and
    optional voice.
    """

    language: Language


@dataclass
class TTSArgsUpdateFrame(ControlFrame):
    """A control frame containing a request to update to TTS args."""

    args: dict


@dataclass
class AvatarModelUpdateFrame(ControlFrame):
    """A control frame containing a request to update the Avatar model and optional
    language.
    """

    model: str


@dataclass
class AvatarLanguageUpdateFrame(ControlFrame):
    """A control frame containing a request to update to Avatar language."""

    language: Language


@dataclass
class AvatarArgsUpdateFrame(ControlFrame):
    """A control frame containing a request to update to Avatar args."""

    args: dict


@dataclass
class IntervalFrame(ControlFrame):
    """window.setInterval."""

    types: List[type] = field(default_factory=[])
    interval_time_ms: int = 0



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from collections import Counter
from multiprocessing import Pool

import torch
from fairseq import utils
from fairseq.binarizer import safe_readline
from fairseq.data import data_utils
from fairseq.file_io import PathManager
from fairseq.tokenizer import tokenize_line
import sentencepiece as spm

class Dictionary:
    """A mapping from symbols to consecutive integers"""

    def __init__(
        self,
        input_file=None,  # begin keyword-only arguments
        spm_model=None,
        bos="<s>",
        pad="<pad>",
        eos="</s>",
        unk="<unk>",
        extra_special_symbols=None,
    ):

        self.symbols = []
        self.count = []
        self.indices = {}
        self.bos_word, self.unk_word, self.pad_word, self.eos_word = bos, unk, pad, eos
        if input_file is not None and 'json' in input_file:
            self.add_from_json(input_file)
        self.bos_index = self.add_symbol(bos)
        self.pad_index = self.add_symbol(pad)
        self.eos_index = self.add_symbol(eos)
        self.unk_index = self.add_symbol(unk)
        if extra_special_symbols:
            for s in extra_special_symbols:
                self.add_symbol(s)
        self.nspecial = len(self.symbols)
        if input_file is not None and 'txt' in input_file:
            self.add_from_file(input_file)            
        else:
            pass
        if spm_model is not None:
            self.processor = spm.SentencePieceProcessor()
            self.processor.load(spm_model)
        else:
            self.processor = None

    def __eq__(self, other):
        return self.indices == other.indices

    def __getitem__(self, idx):
        if idx < len(self.symbols):
            return self.symbols[idx]
        return self.unk_word

    def __len__(self):
        """Returns the number of symbols in the dictionary"""
        return len(self.symbols)

    def __contains__(self, sym):
        return sym in self.indices

    def index(self, sym):
        """Returns the index of the specified symbol"""
        assert isinstance(sym, str)
        if sym in self.indices:
            return self.indices[sym]
        return self.unk_index

    def string(
        self,
        tensor,
        bpe_symbol=None,
        escape_unk=False,
        extra_symbols_to_ignore=None,
        unk_string=None,
        include_eos=False,
        separator=" ",
    ):
        """Helper for converting a tensor of token indices to a string.

        Can optionally remove BPE symbols or escape <unk> words.
        """
        if torch.is_tensor(tensor) and tensor.dim() == 2:
            return "\n".join(
                self.string(t, bpe_symbol, escape_unk, extra_symbols_to_ignore, include_eos=include_eos)
                for t in tensor
            )

        extra_symbols_to_ignore = set(extra_symbols_to_ignore or [])
        extra_symbols_to_ignore.add(self.eos())

        def token_string(i):
            if i == self.unk():
                if unk_string is not None:
                    return unk_string
                else:
                    return self.unk_string(escape_unk)
            else:
                return self[i]

        if hasattr(self, "bos_index"):
            extra_symbols_to_ignore.add(self.bos())

        sent = separator.join(
            token_string(i)
            for i in tensor
            if utils.item(i) not in extra_symbols_to_ignore
        )

        return data_utils.post_process(sent, bpe_symbol)

    def unk_string(self, escape=False):
        """Return unknown string, optionally escaped as: <<unk>>"""
        if escape:
            return "<{}>".format(self.unk_word)
        else:
            return self.unk_word

    def add_symbol(self, word, n=1, overwrite=False):
        """Adds a word to the dictionary"""
        if word in self.indices and not overwrite:
            idx = self.indices[word]
            self.count[idx] = self.count[idx] + n
            return idx
        else:
            idx = len(self.symbols)
            self.indices[word] = idx
            self.symbols.append(word)
            self.count.append(n)
            return idx

    def update(self, new_dict):
        """Updates counts from new dictionary."""
        for word in new_dict.symbols:
            idx2 = new_dict.indices[word]
            if word in self.indices:
                idx = self.indices[word]
                self.count[idx] = self.count[idx] + new_dict.count[idx2]
            else:
                idx = len(self.symbols)
                self.indices[word] = idx
                self.symbols.append(word)
                self.count.append(new_dict.count[idx2])

    def finalize(self, threshold=-1, nwords=-1, padding_factor=8):
        """Sort symbols by frequency in descending order, ignoring special ones.

        Args:
            - threshold defines the minimum word count
            - nwords defines the total number of words in the final dictionary,
                including special symbols
            - padding_factor can be used to pad the dictionary size to be a
                multiple of 8, which is important on some hardware (e.g., Nvidia
                Tensor Cores).
        """
        if nwords <= 0:
            nwords = len(self)

        new_indices = dict(zip(self.symbols[: self.nspecial], range(self.nspecial)))
        new_symbols = self.symbols[: self.nspecial]
        new_count = self.count[: self.nspecial]

        c = Counter(
            dict(
                sorted(zip(self.symbols[self.nspecial :], self.count[self.nspecial :]))
            )
        )
        for symbol, count in c.most_common(nwords - self.nspecial):
            if count >= threshold:
                new_indices[symbol] = len(new_symbols)
                new_symbols.append(symbol)
                new_count.append(count)
            else:
                break

        assert len(new_symbols) == len(new_indices)

        self.count = list(new_count)
        self.symbols = list(new_symbols)
        self.indices = new_indices

        self.pad_to_multiple_(padding_factor)

    def pad_to_multiple_(self, padding_factor):
        """Pad Dictionary size to be a multiple of *padding_factor*."""
        if padding_factor > 1:
            i = 0
            while len(self) % padding_factor != 0:
                symbol = "madeupword{:04d}".format(i)
                self.add_symbol(symbol, n=0)
                i += 1

    def bos(self):
        """Helper to get index of beginning-of-sentence symbol"""
        return self.bos_index

    def pad(self):
        """Helper to get index of pad symbol"""
        return self.pad_index

    def eos(self):
        """Helper to get index of end-of-sentence symbol"""
        return self.eos_index

    def unk(self):
        """Helper to get index of unk symbol"""
        return self.unk_index

    @classmethod
    def load(cls, f):
        """Loads the dictionary from a text file with the format:

        ```
        <symbol0> <count0>
        <symbol1> <count1>
        ...
        ```
        """
        d = cls()
        d.add_from_file(f)
        return d

    @classmethod
    def load_from_json(cls, f):
        d = cls()
        d.add_from_json(f)
        return d

    def add_from_json(self, f):
        import json
        if isinstance(f, str):
            try:
                with open(PathManager.get_local_path(f), "r", encoding="utf-8") as fd:
                    self.add_from_json(fd)
            except FileNotFoundError as fnfe:
                raise fnfe
            except UnicodeError:
                raise Exception(
                    "Incorrect encoding detected in {}, please "
                    "rebuild the dataset".format(f)
                )
            return

        vocab = json.load(f)
        for k, v in vocab.items():
            self.add_symbol(k)
        

    def add_from_file(self, f):
        """
        Loads a pre-existing dictionary from a text file and adds its symbols
        to this instance.
        """
        if isinstance(f, str):
            try:
                with open(PathManager.get_local_path(f), "r", encoding="utf-8") as fd:
                    self.add_from_file(fd)
            except FileNotFoundError as fnfe:
                raise fnfe
            except UnicodeError:
                raise Exception(
                    "Incorrect encoding detected in {}, please "
                    "rebuild the dataset".format(f)
                )
            return

        lines = f.readlines()
        indices_start_line = self._load_meta(lines)

        for line in lines[indices_start_line:]:
            try:
                line, field = line.rstrip().rsplit(" ", 1)
                if field == "#fairseq:overwrite":
                    overwrite = True
                    line, field = line.rsplit(" ", 1)
                else:
                    overwrite = False
                count = int(field)
                word = line
                if word in self:
                    continue
                    raise RuntimeError(
                        "Duplicate word found when loading Dictionary: '{}'. "
                        "Duplicate words can overwrite earlier ones by adding the "
                        "#fairseq:overwrite flag at the end of the corresponding row "
                        "in the dictionary file. If using the Camembert model, please "
                        "download an updated copy of the model file.".format(word)
                    )
                self.add_symbol(word, n=count, overwrite=overwrite)
            except ValueError:
                raise ValueError(
                    "Incorrect dictionary format, expected '<token> <cnt> [flags]'"
                )

    def _save(self, f, kv_iterator):
        if isinstance(f, str):
            PathManager.mkdirs(os.path.dirname(f))
            with PathManager.open(f, "w", encoding="utf-8") as fd:
                return self.save(fd)
        for k, v in kv_iterator:
            print("{} {}".format(k, v), file=f)

    def _get_meta(self):
        return [], []

    def _load_meta(self, lines):
        return 0

    def save(self, f):
        """Stores dictionary into a text file"""
        ex_keys, ex_vals = self._get_meta()
        self._save(
            f,
            zip(
                ex_keys + self.symbols[self.nspecial :],
                ex_vals + self.count[self.nspecial :],
            ),
        )

    def dummy_sentence(self, length):
        t = torch.Tensor(length).uniform_(self.nspecial + 1, len(self)).long()
        t[-1] = self.eos()
        return t

    def encode_line(
        self,
        line,
        line_tokenizer=tokenize_line,
        add_if_not_exist=True,
        consumer=None,
        append_eos=True,
        reverse_order=False,
    ) -> torch.IntTensor:
        if not isinstance(line, list):
            if self.processor is not None:
                words = self.processor.encode_as_pieces(line)
            else:
                words = line_tokenizer(line)
        else:
            words = line
        if reverse_order:
            words = list(reversed(words))
        nwords = len(words)
        ids = torch.IntTensor(nwords + 1 if append_eos else nwords)

        for i, word in enumerate(words):
            if add_if_not_exist:
                idx = self.add_symbol(word)
            else:
                idx = self.index(word)
            if consumer is not None:
                consumer(word, idx)
            ids[i] = idx
        if append_eos:
            ids[nwords] = self.eos_index
        return ids

    @staticmethod
    def _add_file_to_dictionary_single_worker(
        filename, tokenize, eos_word, worker_id=0, num_workers=1
    ):
        counter = Counter()
        with open(PathManager.get_local_path(filename), "r", encoding="utf-8") as f:
            size = os.fstat(f.fileno()).st_size
            chunk_size = size // num_workers
            offset = worker_id * chunk_size
            end = offset + chunk_size
            f.seek(offset)
            if offset > 0:
                safe_readline(f)  # drop first incomplete line
            line = f.readline()
            while line:
                for word in tokenize(line):
                    counter.update([word])
                counter.update([eos_word])
                # f.tell() returns only an opaque number which can
                # return to the position in the file via f.seek()
                # and does not necessarily represent a byte position
                # in the file. However, f.tell() is faithful to the
                # byte position _most of the time_. Thus we can just
                # check against the file size to prevent early exit.
                if f.tell() > end and f.tell() < size:
                    break
                line = f.readline()
        return counter

    @staticmethod
    def add_file_to_dictionary(filename, dict, tokenize, num_workers):
        def merge_result(counter):
            for w, c in sorted(counter.items()):
                dict.add_symbol(w, c)

        if num_workers > 1:
            pool = Pool(processes=num_workers)
            results = []
            for worker_id in range(num_workers):
                results.append(
                    pool.apply_async(
                        Dictionary._add_file_to_dictionary_single_worker,
                        (filename, tokenize, dict.eos_word, worker_id, num_workers),
                    )
                )
            pool.close()
            pool.join()
            for r in results:
                merge_result(r.get())
        else:
            merge_result(
                Dictionary._add_file_to_dictionary_single_worker(
                    filename, tokenize, dict.eos_word
                )
            )


class TruncatedDictionary(object):
    def __init__(self, wrapped_dict, length):
        self.__class__ = type(
            wrapped_dict.__class__.__name__,
            (self.__class__, wrapped_dict.__class__),
            {},
        )
        self.__dict__ = wrapped_dict.__dict__
        self.wrapped_dict = wrapped_dict
        self.length = min(len(self.wrapped_dict), length)

    def __len__(self):
        return self.length

    def __getitem__(self, i):
        if i < self.length:
            return self.wrapped_dict[i]
        return self.wrapped_dict.unk()



from dataclasses import dataclass

from achatbot.common.types import RATE


@dataclass
class ASRArgs:
    download_path: str = ""
    model_name_or_path: str = "base"

    warmup_steps: int = 2
    # asr
    # NOTE:
    # - openai-whisper or whispertimestamped use str(file_path)/np.ndarray/torch tensor
    # - transformers whisper use torch tensor/tf tensor
    # - faster whisper don't use torch tensor, use np.ndarray or str(file_path)/~BinaryIO~
    # - mlx whisper don't use torch tensor, use str(file_path)/np.ndarray/~mlx.array~
    # - funasr whisper, SenseVoiceSmall use str(file_path)/torch tensor
    # asr_audio: str | bytes | IO[bytes] | np.ndarray | torch.Tensor = None

    # https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes
    language: str = "zh"
    verbose: bool = True
    prompt: str = "<|startoftranscript|>"
    sample_rate: int = RATE
    device: str | dict | None = None
    batch_size: int = 1

    # https://docs.pytorch.org/docs/stable/generated/torch.compile.html
    torch_compile_mode: str = "reduce-overhead"
    # triton_cudagraphs: bool = True



"""
FastMCP Echo Server
"""

from fastmcp import Context
from mcp.types import (
    ToolAnnotations,
)

from achatbot.modules.functions.weather.api import WeatherFuncEnvInit
from achatbot.common.session import Session

from . import mcp

# https://gofastmcp.com/servers/tools


# NOTE: dispatch to execute
item = WeatherFuncEnvInit.initWeatherEngine().get_tool_call()
function = item.get("function")
if function is None:

    @mcp.tool(name="get weather", description="get weather")
    def get_weather(longitude: float, latitude: float, ctx: Context) -> str:
        session = Session(client_id=ctx.client_id)
        return WeatherFuncEnvInit.initWeatherEngine().execute(session, longitude, latitude)
else:

    @mcp.tool(
        name=function.get("name"),
        description=function.get("description"),
        annotations=ToolAnnotations(title="weather"),
    )
    def get_weather(longitude: float, latitude: float, ctx: Context) -> str:
        session = Session(client_id=ctx.client_id)
        return WeatherFuncEnvInit.initWeatherEngine().execute(
            session, **{"longitude": longitude, "latitude": latitude}
        )



import time
import logging

from apipeline.frames import Frame, AudioRawFrame
from apipeline.processors.frame_processor import FrameDirection
from typing import Type

from achatbot.common.interface import ISpeechEnhancer
from achatbot.processors.session_processor import SessionProcessor
from achatbot.types.frames import (
    VADStateAudioRawFrame,
    UserStartedSpeakingFrame,
    UserStoppedSpeakingFrame,
)
from achatbot.common.session import Session


class AudioNoiseFilter(SessionProcessor):
    def __init__(
        self,
        se: ISpeechEnhancer,
        start_frame_type: Type[Frame] = UserStartedSpeakingFrame,
        end_frame_type: Type[Frame] = UserStoppedSpeakingFrame,
        session: Session | None = None,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self._start_frame_type = start_frame_type
        self._end_frame_type = end_frame_type
        self._se = se
        self._user_speaking = False
        self._verbose = kwargs.get("verbose", False)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, self._start_frame_type):
            self._user_speaking = True
        if isinstance(frame, self._end_frame_type):
            self._user_speaking = False

        if isinstance(frame, AudioRawFrame):
            if self._user_speaking is False:
                await self.push_frame(frame)
                return

            self.session.ctx.state["audio_chunk"] = frame.audio
            self.session.ctx.state["sample_rate"] = frame.sample_rate
            self.session.ctx.state["is_last"] = False
            if isinstance(frame, VADStateAudioRawFrame):
                self.session.ctx.state["is_last"] = frame.is_final
            start = time.time()
            filter_audio = self._se.enhance(self.session)  # only read, don't CoW
            frame.audio = filter_audio
            if self._verbose:
                print(
                    f"{self.name} filter audio_chunk_len: {len(self.session.ctx.state['audio_chunk'])}"
                    f" {str(frame)} cost: {time.time() - start} s"
                )
        await self.push_frame(frame)



import argparse
import json
import os
from dataclasses import dataclass
from typing import Any, Dict, List, Union

import soundfile as sf
import torch
from datasets import Audio, Dataset
from peft import LoraConfig, PeftModel, get_peft_model, prepare_model_for_kbit_training
from tqdm import tqdm
from transformers import (
    AutoModelForSpeechSeq2Seq,
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
    WhisperForConditionalGeneration,
    WhisperProcessor,
)


def setup(rank, world_size):
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"
    torch.distributed.init_process_group("nccl", rank=rank, world_size=world_size)


def cleanup():
    torch.distributed.destroy_process_group()


@dataclass
class DataCollatorSpeechSeq2SeqWithPadding:
    processor: Any

    def __call__(
        self, features: List[Dict[str, Union[List[int], torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        input_features = [{"input_features": feature["input_features"]} for feature in features]
        label_features = [{"input_ids": feature["labels"]} for feature in features]

        batch = self.processor.feature_extractor.pad(input_features, return_tensors="pt")
        labels_batch = self.processor.tokenizer.pad(label_features, return_tensors="pt")

        labels = labels_batch["input_ids"].masked_fill(labels_batch.attention_mask.ne(1), -100)

        bos_mask = labels[:, 0] == self.processor.tokenizer.bos_token_id
        if bos_mask.all().cpu().item():
            labels = labels[:, 1:]

        batch["labels"] = labels

        return batch


def check_audio_file(file_path):
    try:
        with sf.SoundFile(file_path):
            return True
    except Exception:
        return False


def load_dataset(data_dir, json_file):
    with open(os.path.join(data_dir, json_file)) as f:
        transcriptions = json.load(f)

    valid_data = []
    for k, v in tqdm(transcriptions.items(), desc="Checking audio files"):
        file_path = os.path.join(data_dir, f"{k}.wav")
        if check_audio_file(file_path):
            valid_data.append({"audio": file_path, "sentence": v})

    dataset = Dataset.from_list(valid_data)
    dataset = dataset.cast_column("audio", Audio(sampling_rate=16000))

    return dataset


def prepare_dataset(batch, processor, device):
    audio = batch["audio"]
    input_features = processor.feature_extractor(
        audio["array"], sampling_rate=audio["sampling_rate"]
    ).input_features[0]
    batch["input_features"] = torch.tensor(input_features).to(device)
    batch["labels"] = torch.tensor(processor.tokenizer(batch["sentence"]).input_ids).to(device)

    return batch


class CustomTrainer(Seq2SeqTrainer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_log = []

    def training_step(self, model, inputs):
        loss = super().training_step(model, inputs)
        if self.state.global_step % 50 == 0:
            self.loss_log.append((self.state.global_step, loss.item()))
            print(f"Step {self.state.global_step}: Loss = {loss.item()}")
        return loss


def main(rank, world_size, args):
    setup(rank, world_size)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model = WhisperForConditionalGeneration.from_pretrained(
        args.model_name_or_path, load_in_8bit=True, device_map="auto"
    )
    processor = WhisperProcessor.from_pretrained(
        args.model_name_or_path, language=args.language, task=args.task
    )

    dataset = load_dataset(args.data_dir, args.json_file)
    dataset = dataset.train_test_split(test_size=0.1)

    print("Preparing dataset...")
    dataset = dataset.map(
        lambda batch: prepare_dataset(batch, processor, device),
        remove_columns=dataset["train"].column_names,
        num_proc=args.num_proc,
    )

    model = prepare_model_for_kbit_training(model)
    config = LoraConfig(
        r=32, lora_alpha=64, target_modules=["q_proj", "v_proj"], lora_dropout=0.05, bias="none"
    )
    model = get_peft_model(model, config)

    training_args = Seq2SeqTrainingArguments(
        output_dir=args.output_dir,
        per_device_train_batch_size=args.batch_size,
        gradient_accumulation_steps=args.gradient_accumulation_steps,
        learning_rate=args.learning_rate,
        num_train_epochs=args.num_epochs,
        fp16=True,
        evaluation_strategy="steps",
        per_device_eval_batch_size=args.batch_size,
        logging_steps=100,
        save_steps=1000,
        eval_steps=1000,
        remove_unused_columns=False,
        label_names=["labels"],
        local_rank=rank,
    )

    data_collator = DataCollatorSpeechSeq2SeqWithPadding(processor=processor)

    trainer = CustomTrainer(
        model=model,
        args=training_args,
        train_dataset=dataset["train"],
        eval_dataset=dataset["test"],
        tokenizer=processor.feature_extractor,
        data_collator=data_collator,
    )

    trainer.train()
    trainer.save_model(args.output_dir)

    # Save loss log
    with open(os.path.join(args.output_dir, "loss_log.json"), "w") as f:
        json.dump(trainer.loss_log, f)

    cleanup()


def inference(args):
    base_model = AutoModelForSpeechSeq2Seq.from_pretrained(
        args.model_name_or_path,
        torch_dtype=torch.float16,
        device_map="auto",
    )

    peft_model = PeftModel.from_pretrained(base_model, args.output_dir)
    merged_model = peft_model.merge_and_unload()

    processor = WhisperProcessor.from_pretrained(args.model_name_or_path)

    audio_file = "path/to/your/audio/file.wav"
    audio_input = processor(audio_file, return_tensors="pt").input_features

    generated_ids = merged_model.generate(audio_input)
    transcription = processor.batch_decode(generated_ids, skip_special_tokens=True)[0]
    print(f"Transcription: {transcription}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model_name_or_path", type=str, default="openai/whisper-large-v3")
    parser.add_argument(
        "--data_dir", type=str, required=True, help="Directory containing wav files and json file"
    )
    parser.add_argument(
        "--json_file",
        type=str,
        default="successful_texts.json",
        help="JSON file containing transcriptions",
    )
    parser.add_argument("--language", type=str, default="English")
    parser.add_argument("--task", type=str, default="transcribe")
    parser.add_argument("--output_dir", type=str, default="./whisper_output")
    parser.add_argument("--batch_size", type=int, default=2)
    parser.add_argument("--gradient_accumulation_steps", type=int, default=1)
    parser.add_argument("--learning_rate", type=float, default=1e-3)
    parser.add_argument("--num_epochs", type=int, default=3)
    parser.add_argument("--num_proc", type=int, default=1)
    parser.add_argument("--mode", type=str, choices=["train", "inference"], default="train")
    args = parser.parse_args()

    if args.mode == "train":
        world_size = torch.cuda.device_count()
        torch.multiprocessing.spawn(main, args=(world_size, args), nprocs=world_size, join=True)
    elif args.mode == "inference":
        inference(args)



import logging
from time import perf_counter
import time
import traceback
from threading import Thread
from typing import Dict
from functools import lru_cache

import numpy as np

try:
    import torch

    from qwen_omni_utils import process_mm_info
    from transformers import (
        AutoConfig,
        TextIteratorStreamer,
        Qwen3OmniMoeProcessor,
    )
    from src.core.llm.transformers.models.qwen3_omni import (
        Qwen3OmniMoeForConditionalGenerationStreaming,
    )

except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Qwen3Omni, you need to `pip install achatbot[llm_transformers_manual_vision_voice_qwen]`"
    )
    raise Exception(f"Missing module: {e}")


from src.common.utils.thread_safe import ThreadSafeDict
from src.common.utils.helper import get_device
from src.common.random import set_all_random_seed
from src.common.session import Session
from src.types.omni.qwen3_vision_voice import (
    Qwen3TransformersVisionVoiceLMArgs,
    Qwen3OmniCode2WavArgs,
)
from src.types.llm.transformers import TransformersLMArgs
from src.types.speech.language import TO_LLM_LANGUAGE
from src.common.chat_history import ChatHistory
from .base import TransformersBaseLLM


class TransformersManualQwen3OmniLLM(TransformersBaseLLM):
    """
    vision/speech to text+speech voice chat
    """

    TAG = "llm_transformers_manual_qwen3omni"

    # Voice settings
    RATE = 24000
    SPEAKER_LIST = ["Chelsie", "Ethan"]

    def __init__(self, **args) -> None:
        self.args = Qwen3TransformersVisionVoiceLMArgs()
        self.args.update(**args)
        self.args.lm_device = self.args.lm_device or get_device()
        self.thinker_args = TransformersLMArgs(**self.args.thinker_args)
        self.talker_args = TransformersLMArgs(**self.args.talker_args)
        self.code2wav_args = Qwen3OmniCode2WavArgs(**self.args.code2wav_args)
        logging.info(f"Model args: {args}")
        logging.info(f"Model thinker_args: {self.thinker_args}")
        logging.info(f"Model talker_args: {self.talker_args}")
        logging.info(f"Model code2wav_args: {self.code2wav_args}")
        config = AutoConfig.from_pretrained(self.args.lm_model_name_or_path)
        config.enable_audio_output = True
        if self.args.disable_talker is True:
            config.enable_audio_output = False

        if self.args.lm_device_map:
            self._model: Qwen3OmniMoeForConditionalGenerationStreaming = (
                Qwen3OmniMoeForConditionalGenerationStreaming.from_pretrained(
                    self.args.lm_model_name_or_path,
                    dtype=self.args.lm_torch_dtype,
                    #!NOTE: https://github.com/huggingface/transformers/issues/20896
                    # device_map for multi cpu/gpu with accelerate
                    device_map=self.args.lm_device_map,
                    attn_implementation=self.args.lm_attn_impl,
                    trust_remote_code=True,
                    config=config,
                ).eval()
            )
        else:
            self._model: Qwen3OmniMoeForConditionalGenerationStreaming = (
                Qwen3OmniMoeForConditionalGenerationStreaming.from_pretrained(
                    self.args.lm_model_name_or_path,
                    dtype=self.args.lm_torch_dtype,
                    attn_implementation=self.args.lm_attn_impl,
                    trust_remote_code=True,
                    config=config,
                )
                .eval()
                .to(self.args.lm_device)
            )

        self._tokenizer = Qwen3OmniMoeProcessor.from_pretrained(
            self.args.lm_model_name_or_path,
            use_fast=True,
        )

        self.session_chat_history: Dict[str, ChatHistory] = ThreadSafeDict()

        self.warmup()

    def warmup(self):
        if self.args.warmup_steps <= 0:
            return
        logging.info(
            f"Warming up {self.__class__.__name__} device: {self._model.device} with {self.args.warmup_steps} steps"
        )

        dummy_msgs = [
            {
                "role": self.args.init_chat_role,
                "content": [
                    {
                        "type": "text",
                        "text": self.args.init_chat_prompt or self.CHAT_SYS_PROMPT,
                    }
                ],
            },
            {
                "role": self.args.user_role,
                "content": [
                    {"type": "text", "text": self.args.warmup_prompt or "请简单介绍下自己"}
                ],
            },
        ]
        # Preparation for inference
        text = self._tokenizer.apply_chat_template(
            dummy_msgs, tokenize=False, add_generation_prompt=True
        )
        if self.args.verbose:
            print(f"Warmup prompt token: {text}")
        audios, images, videos = process_mm_info(dummy_msgs, use_audio_in_video=False)
        inputs = self._tokenizer(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=False,
        )
        inputs = inputs.to(self._model.device).to(self._model.dtype)

        if "cuda" in str(self._model.device):
            start_event = torch.cuda.Event(enable_timing=True)
            end_event = torch.cuda.Event(enable_timing=True)
            torch.cuda.synchronize()
            start_event.record()

        for step in range(self.args.warmup_steps):
            streamer = self._model.generate_stream(
                inputs,
                use_audio_in_video=False,
                thinker_max_tokens_per_step=self.thinker_args.lm_gen_max_tokens_per_step,
                thinker_max_new_tokens=15,
                thinker_top_k=self.thinker_args.lm_gen_top_k,
                thinker_top_p=self.thinker_args.lm_gen_top_p,
                thinker_temperature=self.thinker_args.lm_gen_temperature,
                thinker_repetition_penalty=self.thinker_args.lm_gen_repetition_penalty,
                thinker_eos_token_ids=self.args.thinker_eos_token_ids,
                thinker_stop_strings_per_step=self.args.thinker_stop_strings_per_step,
                tokenizer=self._tokenizer.tokenizer,
                return_audio=self._model.has_talker,
                speaker=self.args.speaker,
                talker_top_k=self.talker_args.lm_gen_top_k,
                talker_top_p=self.talker_args.lm_gen_top_p,
                talker_temperature=self.talker_args.lm_gen_temperature,
                talker_repetition_penalty=self.talker_args.lm_gen_repetition_penalty,
                talker_min_new_tokens=self.talker_args.lm_gen_min_new_tokens,
                talker_max_new_tokens=self.talker_args.lm_gen_max_new_tokens,
                token2wav_kwargs=self.code2wav_args.__dict__,
                skip_token_ids=self.skip_token_ids(),
            )
            times = []
            start_time = time.perf_counter()
            text = ""
            for i, chunk in enumerate(streamer):
                times.append(time.perf_counter() - start_time)
                if "thinker_ids" in chunk:
                    text = self._tokenizer.decode(chunk["thinker_ids"][0], skip_special_tokens=True)
                    logging.info(f"{i} chunk: {text} , warmup time: {times[i]} s")
                if "talker_wav" in chunk:
                    logging.info(
                        f"{i} chunk: {text} | {chunk['talker_wav'].shape} , warmup time: {times[i]} s"
                    )
                start_time = time.perf_counter()
            if len(times) > 0:
                logging.info(
                    f"step {step} warmup TTFT(chunk) time: {times[0]} s | total: {sum(times)} s"
                )
            else:
                logging.warning(f"step {step} warmup no generate stream")
            step += 1

        if "cuda" in str(self._model.device):
            end_event.record()
            torch.cuda.synchronize()
            logging.info(
                f"{self.__class__.__name__}:  warmed up! time: {start_event.elapsed_time(end_event) * 1e-3:.3f} s"
            )

    @lru_cache
    def skip_token_ids(self):
        token_ids = []
        for i in ",;.?，；。？！":
            # for i in ",.":
            token_id = self._tokenizer.tokenizer.encode(i)
            token_ids.extend(token_id)
        return token_ids

    def get_prompt(self, session: Session) -> list:
        prompt = []
        if isinstance(session.ctx.state["prompt"], list):
            prompt = session.ctx.state["prompt"]
        return prompt

    @torch.no_grad()
    def generate(self, session: Session, **kwargs):
        """
        - prompt:
        [
            {"type": "image", "image": url / path / base64 / nparray},
            {"type": "video", "video": url / path / base64 / nparray},
            {"type": "audio", "audio": url / path / base64 / nparray},
            {"type": "text", "text": str},
        ]

        - return Generator[dict, None, None]:
        {
            "text": str,
            "audio_wav": torch.Tensor,# (T,)
        }
        """
        seed = kwargs.get("seed", self.args.lm_gen_seed)
        set_all_random_seed(seed)

        prompt = self.get_prompt(session)

        message = {"role": self.args.user_role, "content": prompt}
        self.add_chat_history(session, message)
        messages = self.get_session_chat_history(session.ctx.client_id)
        logging.debug(f"messages: {messages}")

        text = self._tokenizer.apply_chat_template(
            messages, tokenize=False, add_generation_prompt=True
        )
        audios, images, videos = process_mm_info(
            messages, use_audio_in_video=kwargs.get("use_audio_in_video", False)
        )
        if self.args.verbose:
            print(f"prompt token: {text}")
        {
            logging.debug(f"audios[{i}]: {item.shape}") for i, item in enumerate(audios)
        } if audios else logging.debug(audios)
        {
            logging.debug(f"images[{i}]: {item}") for i, item in enumerate(images)
        } if images else logging.debug(images)
        {
            logging.debug(f"videos[{i}]: {item.shape}") for i, item in enumerate(videos)
        } if videos else logging.debug(videos)

        inputs = self._tokenizer(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=kwargs.get("use_audio_in_video", False),
        )
        inputs = inputs.to(self._model.device).to(self._model.dtype)
        for k, v in inputs.items():
            logging.debug(f"{k}: {v.shape}")

        return_audio = kwargs.get("return_audio", self._model.has_talker)
        gen_assistant_text = ""
        try:
            if not return_audio:  # text / vision(image/video) / audio / text + image -> text
                for item in self.thinker_stream(
                    inputs,
                    use_audio_in_video=kwargs.get("use_audio_in_video", False),
                    thinker_top_k=kwargs.get("thinker_top_k", None)
                    or self.thinker_args.lm_gen_top_k,
                    thinker_top_p=kwargs.get("thinker_top_p", None)
                    or self.thinker_args.lm_gen_top_p,
                    thinker_temperature=kwargs.get("thinker_temperature", None)
                    or self.thinker_args.lm_gen_temperature,
                    thinker_repetition_penalty=kwargs.get("thinker_repetition_penalty", None)
                    or self.thinker_args.lm_gen_repetition_penalty,
                    thinker_min_new_tokens=kwargs.get("thinker_min_new_tokens", None)
                    or self.thinker_args.lm_gen_min_new_tokens,
                    thinker_max_new_tokens=kwargs.get("thinker_max_new_tokens", None)
                    or self.thinker_args.lm_gen_max_new_tokens,
                ):
                    gen_assistant_text += item["text"]
                    yield item
            else:  # text / vision(image/video) / audio / text + image -> text + audio
                stream = self._model.generate_stream(
                    inputs,
                    use_audio_in_video=kwargs.get("use_audio_in_video", False),
                    thinker_max_tokens_per_step=kwargs.get("thinker_max_tokens_per_step", None)
                    or self.thinker_args.lm_gen_max_tokens_per_step,
                    thinker_max_new_tokens=kwargs.get("thinker_max_new_tokens", None)
                    or self.thinker_args.lm_gen_max_new_tokens,
                    thinker_top_k=kwargs.get("thinker_top_k", None)
                    or self.thinker_args.lm_gen_top_k,
                    thinker_top_p=kwargs.get("thinker_top_p", None)
                    or self.thinker_args.lm_gen_top_p,
                    thinker_temperature=kwargs.get("thinker_temperature", None)
                    or self.thinker_args.lm_gen_temperature,
                    thinker_repetition_penalty=kwargs.get("thinker_repetition_penalty", None)
                    or self.thinker_args.lm_gen_repetition_penalty,
                    thinker_eos_token_ids=kwargs.get("thinker_eos_token_ids", None)
                    or self.args.thinker_eos_token_ids,
                    thinker_stop_strings_per_step=kwargs.get("thinker_stop_strings_per_step", None)
                    or self.args.thinker_stop_strings_per_step,
                    tokenizer=self._tokenizer.tokenizer,
                    return_audio=kwargs.get("return_audio", self._model.has_talker),
                    speaker=kwargs.get("speaker", None) or self.args.speaker,
                    talker_top_k=kwargs.get("talker_top_k", None) or self.talker_args.lm_gen_top_k,
                    talker_top_p=kwargs.get("talker_top_p", None) or self.talker_args.lm_gen_top_p,
                    talker_temperature=kwargs.get("talker_temperature", None)
                    or self.talker_args.lm_gen_temperature,
                    talker_repetition_penalty=kwargs.get("talker_repetition_penalty", None)
                    or self.talker_args.lm_gen_repetition_penalty,
                    talker_min_new_tokens=kwargs.get("talker_min_new_tokens", None)
                    or self.talker_args.lm_gen_min_new_tokens,
                    talker_max_new_tokens=kwargs.get("talker_max_new_tokens", None)
                    or self.talker_args.lm_gen_max_new_tokens,
                    token2wav_kwargs=kwargs.get("token2wav_kwargs", None)
                    or self.code2wav_args.__dict__,
                    skip_token_ids=self.skip_token_ids(),
                )

                gen_text = ""
                for chunk in stream:
                    if "thinker_ids" in chunk:
                        text = self._tokenizer.decode(
                            chunk["thinker_ids"][0], skip_special_tokens=True
                        )
                        if gen_text != text:
                            gen_text = text
                            gen_assistant_text += text
                        yield {"text": text}

                    if "talker_wav" in chunk:
                        # audio_bytes = (
                        #    (chunk["talker_wav"].float().detach().cpu().numpy() * 32768)
                        #    .astype(np.int16)
                        #    .tobytes()
                        # )
                        yield {"audio_wav": chunk["talker_wav"]}
        except Exception as e:
            tb_str = traceback.format_exc()
            logging.error(f"Exception: {e}; traceback: {tb_str}")

        self.add_chat_history(
            session, {"role": "assistant", "content": [{"text": gen_assistant_text}]}
        )

    @torch.no_grad()
    def thinker_stream(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_top_k: int = 40,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_repetition_penalty: float = 1.05,
        thinker_min_new_tokens: int = 1,
        thinker_max_new_tokens: int = 1024,
    ):
        streamer = TextIteratorStreamer(self._tokenizer, skip_prompt=True, skip_special_tokens=True)

        generation_kwargs = dict(
            **inputs,
            streamer=streamer,
            use_audio_in_video=use_audio_in_video,
            return_audio=False,
            thinker_do_sample=True if thinker_temperature > 0.0 else False,
            temperature=thinker_temperature,
            top_k=thinker_top_k,
            top_p=thinker_top_p,
            repetition_penalty=thinker_repetition_penalty,
            min_new_tokens=thinker_min_new_tokens,
            max_new_tokens=thinker_max_new_tokens,
        )
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        times = []
        start_time = perf_counter()
        for new_text in streamer:
            times.append(perf_counter() - start_time)
            generated_text += new_text
            if new_text == "":
                continue
            yield {"text": new_text}
            start_time = perf_counter()

        logging.info(
            f"thinker generate [{generated_text}] TTFT: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
        )


class TransformersManualVisionVoiceQwen3OmniLLM(TransformersManualQwen3OmniLLM):
    """
    vision + speech to speech voice chat

    - vision + speech -> text + speech
    """

    TAG = "llm_transformers_manual_qwen3omni_vision_voice"
    USER_SYSTEM_PROMPT = "You are Qwen-Omni, a smart voice assistant created by Alibaba Qwen."
    CHAT_SYS_PROMPT = f"{USER_SYSTEM_PROMPT} You are a virtual voice assistant with no gender or age.\nYou are communicating with the user.\nIn user messages, “I/me/my/we/our” refer to the user and “you/your” refer to the assistant. In your replies, address the user as “you/your” and yourself as “I/me/my”; never mirror the user’s pronouns—always shift perspective. Keep original pronouns only in direct quotes; if a reference is unclear, ask a brief clarifying question.\nInteract with users using short(no more than 50 words), brief, straightforward language, maintaining a natural tone.\nNever use formal phrasing, mechanical expressions, bullet points, overly structured language. \nYour output must consist only of the spoken content you want the user to hear. \nDo not include any descriptions of actions, emotions, sounds, or voice changes. \nDo not use asterisks, brackets, parentheses, or any other symbols to indicate tone or actions. \nYou must answer users' audio or text questions, do not directly describe the video content. \nYou should communicate in the same language strictly as the user unless they request otherwise.\nWhen you are uncertain (e.g., you can't see/hear clearly, don't understand, or the user makes a comment rather than asking a question), use appropriate questions to guide the user to continue the conversation.\nKeep replies concise and conversational, as if talking face-to-face."

    def __init__(self, **args) -> None:
        args["disable_talker"] = False
        language = "English"
        if args.get("lm_language_code", "") in TO_LLM_LANGUAGE:
            language = TO_LLM_LANGUAGE[args.get("lm_language_code", "")]
        self.CHAT_SYS_PROMPT += f"\nPlease reply to my message in {language}."
        super().__init__(**args)



import cv2
import os
import datetime
import time
from flask import Flask, Response, request

app = Flask(__name__)


def generate_frames_from_video(video_source):
    """
    Generator function to read frames from a video source (path or URL) and yield them as JPEG bytes.
    The frames are yielded at a rate consistent with the original video's FPS.
    """
    cap = cv2.VideoCapture(video_source)
    if not cap.isOpened():
        print(f"Error: Could not open video source {video_source}")
        return

    # 获取视频的帧率
    fps = cap.get(cv2.CAP_PROP_FPS)
    # 获取视频的总帧数
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    print(f"{fps=},{total_frames=}")

    # Calculate the delay needed between frames to match the original video's FPS
    # If fps is 0 (e.g., for some image streams), avoid division by zero
    delay_s = 1.0 / fps if fps > 0 else 0

    while True:
        ret, frame = cap.read()
        if not ret:
            break

        # Encode the frame as JPEG
        height, width, channel = frame.shape
        print(f"Frame dimensions: {width=}x{height=} {channel=}")
        # OpenCV captures frames in BGR format by default.
        # If the intention is to convert to RGB before encoding, perform the conversion.
        # This check assumes 3-channel images from video capture are BGR and need conversion to RGB.
        if len(frame.shape) == 3 and frame.shape[2] == 3:  # Check if it's a 3-channel image
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), 90]  # JPEG quality 90
        _, buffer = cv2.imencode(".jpg", frame, encode_param)
        frame_bytes = buffer.tobytes()

        # Yield the frame in multipart/x-mixed-replace format
        yield (b"--frame\r\nContent-Type: image/jpeg\r\n\r\n" + frame_bytes + b"\r\n")

        # Introduce a delay to match the video's original FPS
        if delay_s > 0:
            time.sleep(delay_s)

    cap.release()
    print(f"Finished streaming frames from {video_source}")


@app.route("/stream_video_frames")
def stream_video_frames():
    """
    HTTP endpoint to stream video frames as a multipart/x-mixed-replace response.
    The 'video_source' parameter can be a local file path or an HTTP URL to a video file.
    Usage: http://<server_ip>:<port>/stream_video_frames?video_source=<path_or_url_to_video>
    """
    video_source = request.args.get("video_source")

    if not video_source:
        return Response("Error: 'video_source' parameter is missing.", status=400)

    # If it's a local file path, check if it exists
    if not video_source.startswith(("http://", "https://")) and not os.path.exists(video_source):
        return Response(f"Error: Local video file not found at {video_source}", status=404)

    return Response(
        generate_frames_from_video(video_source),
        mimetype="multipart/x-mixed-replace; boundary=frame",
    )


# The original video2Images function is kept for completeness if saving to disk is still needed
def video2Images(video_path: str, output_dir: str):
    # 打开视频文件
    cap = cv2.VideoCapture(video_path)

    # 获取视频的帧率
    fps = cap.get(cv2.CAP_PROP_FPS)
    # 获取视频的总帧数
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    print(f"{fps=},{total_frames=}")

    # 创建输出目录
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # 读取视频帧
    frame_count = 0
    while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            break

        # 生成输出文件名
        filename = os.path.join(output_dir, f"frame_{frame_count:06d}.jpg")

        # 保存帧
        cv2.imwrite(filename, frame)
        frame_count += 1

    # 释放资源
    cap.release()
    print(f"Extracted {frame_count} frames from {video_path}")


if __name__ == "__main__":
    # Example for saving images (original functionality)
    # video_path_save = "./videos/cv/capture4_2024-09-15_20-10-29.mp4"
    # output_dir_save = "./images/cv_saved"
    # if os.path.exists(video_path_save):
    #     video2Images(video_path_save, output_dir_save)
    # else:
    #     print(f"Video for saving not found: {video_path_save}. Skipping image extraction.")

    # Run the Flask app for streaming
    # To test, ensure 'videos/cv/capture4_2024-09-15_20-10-29.mp4' exists relative to this script.
    # Then open your browser to http://127.0.0.1:8001/stream_video_frames?video_source=videos/cv/capture4_2024-09-15_20-10-29.mp4
    # To stream from an HTTP URL (e.g., a public MP4 file):
    # http://127.0.0.1:8001/stream_video_frames?video_source=https://www.learningcontainer.com/wp-content/uploads/2020/05/sample-mp4-file.mp4
    app.run(host="0.0.0.0", port=8001, debug=True)



import asyncio
from dataclasses import dataclass
import json
import logging
import os
import shutil
from contextlib import AsyncExitStack
import traceback
from typing import Any, List

import httpx
from dotenv import load_dotenv
from mcp import ClientSession, StdioServerParameters, types
from mcp.client.stdio import stdio_client
from mcp.client.session_group import SseServerParameters, StreamableHttpParameters
from mcp.client.sse import sse_client
from mcp.client.streamable_http import streamablehttp_client

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")


class Configuration:
    """Manages configuration and environment variables for the MCP client."""

    def __init__(self) -> None:
        """Initialize configuration with environment variables."""
        self.load_env()
        self.api_key = os.getenv("TOGETHER_API_KEY")

    @staticmethod
    def load_env() -> None:
        """Load environment variables from .env file."""
        load_dotenv()

    @staticmethod
    def load_config(file_path: str) -> dict[str, Any]:
        """Load server configuration from JSON file.

        Args:
            file_path: Path to the JSON configuration file.

        Returns:
            Dict containing server configuration.

        Raises:
            FileNotFoundError: If configuration file doesn't exist.
            JSONDecodeError: If configuration file is invalid JSON.
        """
        with open(file_path, "r") as f:
            return json.load(f)

    @property
    def llm_api_key(self) -> str:
        """Get the LLM API key.

        Returns:
            The API key as a string.

        Raises:
            ValueError: If the API key is not found in environment variables.
        """
        if not self.api_key:
            raise ValueError("LLM_API_KEY not found in environment variables")
        return self.api_key


class Tool:
    """Represents a tool with its properties and formatting."""

    def __init__(self, name: str, description: str, input_schema: dict[str, Any]) -> None:
        self.name: str = name
        self.description: str = description
        self.input_schema: dict[str, Any] = input_schema

    def format_for_llm(self) -> str:
        """Format tool information for LLM.

        Returns:
            A formatted string describing the tool.
        """
        args_desc = []
        if "properties" in self.input_schema:
            for param_name, param_info in self.input_schema["properties"].items():
                arg_desc = f"- {param_name}: {param_info.get('description', 'No description')}"
                if param_name in self.input_schema.get("required", []):
                    arg_desc += " (required)"
                args_desc.append(arg_desc)

        return f"""
Tool: {self.name}
Description: {self.description}
Arguments:
{chr(10).join(args_desc)}
"""


class Server:
    """Manages MCP server connections and tool execution."""

    def __init__(self, name: str, config: dict[str, Any]) -> None:
        self.name: str = name
        self.config: dict[str, Any] = config
        self.stdio_context: Any | None = None
        self.session: ClientSession | None = None
        self._cleanup_lock: asyncio.Lock = asyncio.Lock()
        self.exit_stack: AsyncExitStack = AsyncExitStack()

    async def initialize(self) -> None:
        """Initialize the server connection."""
        transport = self.config.get("transport", "stdio")
        parameters = self.config.get("parameters", {})
        assert transport in ["stdio", "sse", "streamable-http"]
        if transport == "stdio":
            await self.initialize_stdio(parameters)
        elif transport == "sse":
            await self.initialize_sse(parameters)
        elif transport == "streamable-http":
            await self.initialize_streamable_http(parameters)

    async def initialize_stdio(self, parameters) -> None:
        """Initialize the stdio server connection."""
        server_params = StdioServerParameters(**parameters)
        server_params.env = (
            {**os.environ, **server_params.env} if server_params.env else {**os.environ}
        )
        try:
            transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
            read, write = transport
            session = await self.exit_stack.enter_async_context(ClientSession(read, write))
            await session.initialize()
            self.session = session
        except Exception as e:
            print(e.with_traceback())
            logging.error(f"Error initializing server {self.name}: {e}")
            await self.cleanup()
            raise

    async def initialize_sse(self, parameters) -> None:
        """Initialize the sse server connection."""
        server_params = SseServerParameters(**parameters)
        try:
            transport = await self.exit_stack.enter_async_context(
                sse_client(
                    url=server_params.url,
                    headers=server_params.headers,
                    timeout=server_params.timeout,
                    sse_read_timeout=server_params.sse_read_timeout,
                    auth=None,
                )
            )
            read, write = transport
            session = await self.exit_stack.enter_async_context(ClientSession(read, write))
            await session.initialize()
            self.session = session
        except Exception as e:
            print(e.with_traceback())
            logging.error(f"Error initializing server {self.name}: {e}")
            await self.cleanup()
            raise

    async def initialize_streamable_http(self, parameters) -> None:
        """Initialize the streamable http server connection."""
        server_params = StreamableHttpParameters(**parameters)
        try:
            transport = await self.exit_stack.enter_async_context(
                streamablehttp_client(
                    url=server_params.url,
                    headers=server_params.headers,
                    timeout=server_params.timeout,
                    sse_read_timeout=server_params.sse_read_timeout,
                    terminate_on_close=server_params.terminate_on_close,
                    auth=None,
                )
            )
            read, write, get_session_id = transport
            session = await self.exit_stack.enter_async_context(ClientSession(read, write))
            await session.initialize()
            self.session = session
        except Exception as e:
            traceback.print_exc()
            logging.error(f"Error initializing server {self.name}: {e}")
            await self.cleanup()
            raise

    async def list_tools(self) -> List[Tool]:
        """List available tools from the server.

        Returns:
            A list of available tools.

        Raises:
            RuntimeError: If the server is not initialized.
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        tools_response = await self.session.list_tools()
        tools = []

        for item in tools_response:
            if isinstance(item, tuple) and item[0] == "tools":
                tools.extend(
                    Tool(tool.name, tool.description, tool.inputSchema) for tool in item[1]
                )

        return tools

    async def execute_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        retries: int = 2,
        delay: float = 1.0,
    ) -> Any:
        """Execute a tool with retry mechanism.

        Args:
            tool_name: Name of the tool to execute.
            arguments: Tool arguments.
            retries: Number of retry attempts.
            delay: Delay between retries in seconds.

        Returns:
            Tool execution result.

        Raises:
            RuntimeError: If server is not initialized.
            Exception: If tool execution fails after all retries.
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        attempt = 0
        while attempt < retries:
            try:
                logging.info(f"Executing {tool_name}...")
                result = await self.session.call_tool(tool_name, arguments)

                return result

            except Exception as e:
                attempt += 1
                logging.warning(f"Error executing tool: {e}. Attempt {attempt} of {retries}.")
                if attempt < retries:
                    logging.info(f"Retrying in {delay} seconds...")
                    await asyncio.sleep(delay)
                else:
                    logging.error("Max retries reached. Failing.")
                    raise

    async def cleanup(self) -> None:
        """Clean up server resources."""
        async with self._cleanup_lock:
            try:
                await self.exit_stack.aclose()
                self.session = None
                self.stdio_context = None
            except Exception as e:
                logging.error(f"Error during cleanup of server {self.name}: {e}")


class LLMClient:
    """Manages communication with the LLM provider."""

    def __init__(self, api_key: str) -> None:
        self.api_key: str = api_key

    def get_response(self, messages: list[dict[str, str]]) -> str:
        """Get a response from the LLM.

        Args:
            messages: A list of message dictionaries.

        Returns:
            The LLM's response as a string.

        Raises:
            httpx.RequestError: If the request to the LLM fails.
        """
        url = "https://api.together.xyz/v1/chat/completions"

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        payload = {
            "messages": messages,
            "model": "Qwen/Qwen2.5-7B-Instruct-Turbo",
            "temperature": 0.7,
            "max_tokens": 4096,
            "top_p": 1,
            "stream": False,
            "stop": None,
        }

        try:
            with httpx.Client() as client:
                response = client.post(url, headers=headers, json=payload)
                response.raise_for_status()
                data = response.json()
                return data["choices"][0]["message"]["content"]

        except httpx.RequestError as e:
            error_message = f"Error getting LLM response: {str(e)}"
            logging.error(error_message)

            if isinstance(e, httpx.HTTPStatusError):
                status_code = e.response.status_code
                logging.error(f"Status code: {status_code}")
                logging.error(f"Response details: {e.response.text}")

            return (
                f"I encountered an error: {error_message}. "
                "Please try again or rephrase your request."
            )


class ChatSession:
    """Orchestrates the interaction between user, LLM, and tools."""

    def __init__(self, servers: list[Server], llm_client: LLMClient) -> None:
        self.servers: list[Server] = servers
        self.llm_client: LLMClient = llm_client

    async def cleanup_servers(self) -> None:
        """Clean up all servers properly."""
        for server in reversed(self.servers):
            try:
                await server.cleanup()
            except Exception as e:
                logging.warning(f"Warning during final cleanup: {e}")

    async def process_llm_response(self, llm_response: str) -> str:
        """Process the LLM response and execute tools if needed.

        Args:
            llm_response: The response from the LLM.

        Returns:
            The result of tool execution or the original response.
        """
        import json

        try:
            tool_call = json.loads(llm_response)
            if "tool" in tool_call and "arguments" in tool_call:
                logging.info(f"Executing tool: {tool_call['tool']}")
                logging.info(f"With arguments: {tool_call['arguments']}")

                for server in self.servers:
                    tools = await server.list_tools()
                    if any(tool.name == tool_call["tool"] for tool in tools):
                        try:
                            result = await server.execute_tool(
                                tool_call["tool"], tool_call["arguments"]
                            )

                            if isinstance(result, dict) and "progress" in result:
                                progress = result["progress"]
                                total = result["total"]
                                percentage = (progress / total) * 100
                                logging.info(f"Progress: {progress}/{total} ({percentage:.1f}%)")

                            return f"Tool execution result: {result}"
                        except Exception as e:
                            error_msg = f"Error executing tool: {str(e)}"
                            logging.error(error_msg)
                            return error_msg

                return f"No server found with tool: {tool_call['tool']}"
            return llm_response
        except json.JSONDecodeError:
            return llm_response

    async def start(self) -> None:
        """Main chat session handler."""
        try:
            for server in self.servers:
                try:
                    await server.initialize()
                except Exception as e:
                    logging.error(f"Failed to initialize server: {e}")
                    await self.cleanup_servers()
                    return

            all_tools = []
            for server in self.servers:
                tools = await server.list_tools()
                all_tools.extend(tools)

            tools_description = "\n".join([tool.format_for_llm() for tool in all_tools])

            system_message = (
                "You are a helpful assistant with access to these tools:\n\n"
                f"{tools_description}\n"
                "Choose the appropriate tool based on the user's question. "
                "If no tool is needed, reply directly.\n\n"
                "IMPORTANT: When you need to use a tool, you must ONLY respond with "
                "the exact JSON object format below, nothing else:\n"
                "{\n"
                '    "tool": "tool-name",\n'
                '    "arguments": {\n'
                '        "argument-name": "value"\n'
                "    }\n"
                "}\n\n"
                "After receiving a tool's response:\n"
                "1. Transform the raw data into a natural, conversational response\n"
                "2. Keep responses concise but informative\n"
                "3. Focus on the most relevant information\n"
                "4. Use appropriate context from the user's question\n"
                "5. Avoid simply repeating the raw data\n\n"
                "Please use only the tools that are explicitly defined above."
            )

            messages = [{"role": "system", "content": system_message}]

            while True:
                try:
                    print("---" * 20)
                    user_input = input("You: ").strip().lower()
                    if user_input in ["quit", "exit"]:
                        logging.info("\nExiting...")
                        break
                    if user_input in ["dump", "history"]:
                        print("Dumping history...")
                        print(json.dumps(messages, indent=2))
                        continue
                    if user_input in ["clean", "clear"]:
                        print("Clearing history...")
                        messages.clear()
                        messages = [{"role": "system", "content": system_message}]
                        continue

                    messages.append({"role": "user", "content": user_input})

                    llm_response = self.llm_client.get_response(messages)
                    logging.info("\nAssistant: %s", llm_response)

                    result = await self.process_llm_response(llm_response)

                    if result != llm_response:
                        messages.append({"role": "assistant", "content": llm_response})

                        # NOTE: diff llm use diff function call data to train, so need adapter to standardize as openai format
                        # messages.append({"role": "system", "content": result})
                        messages.append({"role": "tool", "content": result})

                        final_response = self.llm_client.get_response(messages)
                        # TODO fix: neec check result != llm_response again, maybe response tool to call
                        logging.info("\nFinal response: %s", final_response)
                        messages.append({"role": "assistant", "content": final_response})
                    else:
                        messages.append({"role": "assistant", "content": llm_response})

                except KeyboardInterrupt:
                    logging.info("\nExiting...")
                    break

        finally:
            await self.cleanup_servers()


async def main() -> None:
    """Initialize and run the chat session."""
    config = Configuration()
    current_dir = os.path.dirname(os.path.realpath(__file__))
    config_file = os.path.join(current_dir, "servers_config.json")
    server_config = config.load_config(config_file)
    servers = [Server(name, srv_config) for name, srv_config in server_config["mcpServers"].items()]
    llm_client = LLMClient(config.llm_api_key)
    chat_session = ChatSession(servers, llm_client)
    await chat_session.start()


if __name__ == "__main__":
    asyncio.run(main())



import logging

from PIL import Image, ImageOps, ImageDraw, ImageFont


def load_image(image_path) -> Image.Image:
    try:
        image = image_path
        if not isinstance(image_path, Image.Image):
            image = Image.open(image_path)
        corrected_image = ImageOps.exif_transpose(image)
        return corrected_image
    except Exception as e:
        logging.error(f"error: {e}")
        try:
            return Image.open(image_path)
        except Exception:
            return None



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import os


def get_parser():
    parser = argparse.ArgumentParser()
    
    parser.add_argument(
        'input', 
        type=str,
        help="input .tsv file"
    )
    parser.add_argument(
        '--dest',
        type=str,
        help="output directory"
    )

    return parser


def main(args):
    wav_names = []
    text = []
    with open(args.input) as f:
        f.readline()
        for line in f:
            items = line.strip().split("\t")
            wav_names.append(items[1])
            text.append(items[2])
    base_name = os.path.basename(args.input)
    file_name = os.path.splitext(base_name)[0]

    with open(os.path.join(args.dest, file_name+'.list'), 'w') as f:
        for name in wav_names:
            f.write(name+"\n")

    with open(os.path.join(args.dest, file_name+'.text'), 'w') as f:
        for i in range(len(wav_names)):
            f.write("{}\t{}\n".format(wav_names[i], text[i]))

	
if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(args)



from typing import Callable, Optional, Tuple

import torch
import torch.utils.checkpoint
from torch import nn

from transformers.cache_utils import Cache
from transformers.modeling_flash_attention_utils import FlashAttentionKwargs
from transformers.modeling_utils import ALL_ATTENTION_FUNCTIONS
from transformers.processing_utils import Unpack
from transformers.utils import logging
from transformers.models.llama.modeling_llama import (
    LlamaAttention,
    LlamaDecoderLayer,
    LlamaForCausalLM,
    LlamaForQuestionAnswering,
    LlamaForSequenceClassification,
    LlamaForTokenClassification,
    LlamaMLP,
    LlamaModel,
    apply_rotary_pos_emb,
    eager_attention_forward,
)
from .configuration_qwen2 import Qwen2Config


logger = logging.get_logger(__name__)


class Qwen2MLP(LlamaMLP):
    def __init__(self, config):
        super().__init__(config)
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)


class Qwen2Attention(LlamaAttention):
    def __init__(self, config: Qwen2Config, layer_idx: int):
        super().__init__(config, layer_idx)
        self.q_proj = nn.Linear(config.hidden_size, config.num_attention_heads * self.head_dim, bias=True)
        self.k_proj = nn.Linear(config.hidden_size, config.num_key_value_heads * self.head_dim, bias=True)
        self.v_proj = nn.Linear(config.hidden_size, config.num_key_value_heads * self.head_dim, bias=True)
        self.o_proj = nn.Linear(config.num_attention_heads * self.head_dim, config.hidden_size, bias=False)

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor],
        past_key_value: Optional[Cache] = None,
        cache_position: Optional[torch.LongTensor] = None,
        **kwargs: Unpack[FlashAttentionKwargs],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        input_shape = hidden_states.shape[:-1]
        hidden_shape = (*input_shape, -1, self.head_dim)

        query_states = self.q_proj(hidden_states).view(hidden_shape).transpose(1, 2)
        key_states = self.k_proj(hidden_states).view(hidden_shape).transpose(1, 2)
        value_states = self.v_proj(hidden_states).view(hidden_shape).transpose(1, 2)

        cos, sin = position_embeddings
        query_states, key_states = apply_rotary_pos_emb(query_states, key_states, cos, sin)

        if past_key_value is not None:
            # sin and cos are specific to RoPE models; cache_position needed for the static cache
            cache_kwargs = {"sin": sin, "cos": cos, "cache_position": cache_position}
            key_states, value_states = past_key_value.update(key_states, value_states, self.layer_idx, cache_kwargs)

        sliding_window = None
        if (
            self.config.use_sliding_window
            and getattr(self.config, "sliding_window", None) is not None
            and self.layer_idx >= self.config.max_window_layers
        ):
            sliding_window = self.config.sliding_window

        attention_interface: Callable = eager_attention_forward
        if self.config._attn_implementation != "eager":
            if self.config._attn_implementation == "sdpa" and kwargs.get("output_attentions", False):
                logger.warning_once(
                    "`torch.nn.functional.scaled_dot_product_attention` does not support `output_attentions=True`. Falling back to "
                    'eager attention. This warning can be removed using the argument `attn_implementation="eager"` when loading the model.'
                )
            else:
                attention_interface = ALL_ATTENTION_FUNCTIONS[self.config._attn_implementation]

        attn_output, attn_weights = attention_interface(
            self,
            query_states,
            key_states,
            value_states,
            attention_mask,
            dropout=0.0 if not self.training else self.attention_dropout,
            scaling=self.scaling,
            sliding_window=sliding_window,  # main diff with Llama
            **kwargs,
        )

        attn_output = attn_output.reshape(*input_shape, -1).contiguous()
        attn_output = self.o_proj(attn_output)
        return attn_output, attn_weights


class Qwen2DecoderLayer(LlamaDecoderLayer):
    def __init__(self, config: Qwen2Config, layer_idx: int):
        super().__init__()
        self.self_attn = Qwen2Attention(config=config, layer_idx=layer_idx)
        self.mlp = Qwen2MLP(config)
        if config.sliding_window and config._attn_implementation != "flash_attention_2":
            logger.warning_once(
                f"Sliding Window Attention is enabled but not implemented for `{config._attn_implementation}`; "
                "unexpected results may be encountered."
            )


class Qwen2Model(LlamaModel):
    pass


class Qwen2ForCausalLM(LlamaForCausalLM):
    pass


class Qwen2ForSequenceClassification(LlamaForSequenceClassification):
    pass


class Qwen2ForTokenClassification(LlamaForTokenClassification):
    pass


class Qwen2ForQuestionAnswering(LlamaForQuestionAnswering):
    pass



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import os
import warnings
import shutil
import logging

from transformers import AutoTokenizer, AutoModelForCausalLM, AutoConfig, BitsAndBytesConfig
import torch
from llava.model import *
from llava.constants import DEFAULT_IMAGE_PATCH_TOKEN, DEFAULT_IM_START_TOKEN, DEFAULT_IM_END_TOKEN


def load_pretrained_model(model_path, model_base, model_name, load_8bit=False, load_4bit=False, device_map="auto", device="cuda", use_flash_attn=False, **kwargs):
    kwargs = {"device_map": device_map, **kwargs}

    if device != "cuda":
        kwargs['device_map'] = {"": device}

    if load_8bit is True:
        kwargs['load_in_8bit'] = True
    elif load_4bit is True:
        kwargs['load_in_4bit'] = True
        kwargs['quantization_config'] = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=torch.float16,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type='nf4'
        )
    else:
        kwargs['torch_dtype'] = torch.float16

    if use_flash_attn:
        kwargs['attn_implementation'] = 'flash_attention_2'

    logging.info(f"kwargs: {kwargs}")
    if 'llava' in model_name.lower():
        # Load LLaVA model
        if 'lora' in model_name.lower() and model_base is None:
            warnings.warn('There is `lora` in model name but no `model_base` is provided. If you are loading a LoRA model, please provide the `model_base` argument. Detailed instruction: https://github.com/haotian-liu/LLaVA#launch-a-model-worker-lora-weights-unmerged.')
        if 'lora' in model_name.lower() and model_base is not None:
            from llava.model.language_model.llava_llama import LlavaConfig
            lora_cfg_pretrained = LlavaConfig.from_pretrained(model_path)
            tokenizer = AutoTokenizer.from_pretrained(model_base, use_fast=False)
            print('Loading LLaVA from base model...')
            model = LlavaLlamaForCausalLM.from_pretrained(model_base, low_cpu_mem_usage=True, config=lora_cfg_pretrained, **kwargs)
            token_num, tokem_dim = model.lm_head.out_features, model.lm_head.in_features
            if model.lm_head.weight.shape[0] != token_num:
                model.lm_head.weight = torch.nn.Parameter(torch.empty(token_num, tokem_dim, device=model.device, dtype=model.dtype))
                model.model.embed_tokens.weight = torch.nn.Parameter(torch.empty(token_num, tokem_dim, device=model.device, dtype=model.dtype))

            print('Loading additional LLaVA weights...')
            if os.path.exists(os.path.join(model_path, 'non_lora_trainables.bin')):
                non_lora_trainables = torch.load(os.path.join(model_path, 'non_lora_trainables.bin'), map_location='cpu')
            else:
                # this is probably from HF Hub
                from huggingface_hub import hf_hub_download

                def load_from_hf(repo_id, filename, subfolder=None):
                    cache_file = hf_hub_download(
                        repo_id=repo_id,
                        filename=filename,
                        subfolder=subfolder)
                    return torch.load(cache_file, map_location='cpu')
                non_lora_trainables = load_from_hf(model_path, 'non_lora_trainables.bin')
            non_lora_trainables = {(k[11:] if k.startswith('base_model.') else k): v for k, v in non_lora_trainables.items()}
            if any(k.startswith('model.model.') for k in non_lora_trainables):
                non_lora_trainables = {(k[6:] if k.startswith('model.') else k): v for k, v in non_lora_trainables.items()}
            model.load_state_dict(non_lora_trainables, strict=False)

            from peft import PeftModel
            print('Loading LoRA weights...')
            model = PeftModel.from_pretrained(model, model_path)
            print('Merging LoRA weights...')
            model = model.merge_and_unload()
            print('Model is loaded...')
        elif model_base is not None:
            # this may be mm projector only
            print('Loading LLaVA from base model...')
            if 'mpt' in model_name.lower():
                if not os.path.isfile(os.path.join(model_path, 'configuration_mpt.py')):
                    shutil.copyfile(os.path.join(model_base, 'configuration_mpt.py'), os.path.join(model_path, 'configuration_mpt.py'))
                tokenizer = AutoTokenizer.from_pretrained(model_base, use_fast=True)
                cfg_pretrained = AutoConfig.from_pretrained(model_path, trust_remote_code=True)
                model = LlavaMptForCausalLM.from_pretrained(model_base, low_cpu_mem_usage=True, config=cfg_pretrained, **kwargs)
            else:
                tokenizer = AutoTokenizer.from_pretrained(model_base, use_fast=False)
                cfg_pretrained = AutoConfig.from_pretrained(model_path)
                # model = LlavaLlamaForCausalLM.from_pretrained(model_base, low_cpu_mem_usage=True, config=cfg_pretrained, **kwargs)
                model = LlavaQwen2ForCausalLM.from_pretrained(model_base, low_cpu_mem_usage=True, config=cfg_pretrained, **kwargs)

            mm_projector_weights = torch.load(os.path.join(model_path, 'mm_projector.bin'), map_location='cpu')
            mm_projector_weights = {k: v.to(torch.float16) for k, v in mm_projector_weights.items()}
            model.load_state_dict(mm_projector_weights, strict=False)
        else:
            if 'mpt' in model_name.lower():
                tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=True)
                model = LlavaMptForCausalLM.from_pretrained(model_path, low_cpu_mem_usage=True, **kwargs)
            elif 'mistral' in model_name.lower():
                tokenizer = AutoTokenizer.from_pretrained(model_path)
                model = LlavaMistralForCausalLM.from_pretrained(
                    model_path,
                    low_cpu_mem_usage=True,
                    **kwargs
                )
            elif 'dclm' in model_name.lower():
                tokenizer = AutoTokenizer.from_pretrained(model_path)
                model = LlavaOpenlmForCausalLM.from_pretrained(
                    model_path,
                    low_cpu_mem_usage=True,
                    **kwargs
                )
            else:
                tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=False)
                # model = LlavaLlamaForCausalLM.from_pretrained(
                #     model_path,
                #     low_cpu_mem_usage=True,
                #     **kwargs
                # )
                model = LlavaQwen2ForCausalLM.from_pretrained(
                    model_path,
                    low_cpu_mem_usage=True,
                    **kwargs
                )
    else:
        # Load language model
        if model_base is not None:
            # PEFT model
            from peft import PeftModel
            tokenizer = AutoTokenizer.from_pretrained(model_base, use_fast=False)
            model = AutoModelForCausalLM.from_pretrained(model_base, low_cpu_mem_usage=True, **kwargs)
            print(f"Loading LoRA weights from {model_path}")
            model = PeftModel.from_pretrained(model, model_path)
            print(f"Merging weights")
            model = model.merge_and_unload()
            print('Convert to FP16...')
            model.to(torch.float16)
        else:
            use_fast = False
            if 'mpt' in model_name.lower():
                tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=True)
                model = AutoModelForCausalLM.from_pretrained(model_path, low_cpu_mem_usage=True, trust_remote_code=True, **kwargs)
            else:
                tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=False)
                model = AutoModelForCausalLM.from_pretrained(model_path, low_cpu_mem_usage=True, **kwargs)

    image_processor = None

    if 'llava' in model_name.lower():
        mm_use_im_start_end = getattr(model.config, "mm_use_im_start_end", False)
        mm_use_im_patch_token = getattr(model.config, "mm_use_im_patch_token", True)
        if mm_use_im_patch_token:
            tokenizer.add_tokens([DEFAULT_IMAGE_PATCH_TOKEN], special_tokens=True)
        if mm_use_im_start_end:
            tokenizer.add_tokens([DEFAULT_IM_START_TOKEN, DEFAULT_IM_END_TOKEN], special_tokens=True)
        model.resize_token_embeddings(len(tokenizer))

        vision_tower = model.get_vision_tower()
        if not vision_tower.is_loaded:
            vision_tower.load_model(device_map=device_map)
        if device_map != 'auto':
            vision_tower.to(device=device_map, dtype=torch.float16)
        image_processor = vision_tower.image_processor

    if hasattr(model.config, "max_sequence_length"):
        context_len = model.config.max_sequence_length
    else:
        context_len = 2048

    return tokenizer, model, image_processor, context_len



import os
import io
import uuid
import time
import asyncio
import requests
import subprocess
from time import perf_counter
from threading import Thread


import modal

APP_NAME = os.getenv("APP_NAME", "")

app = modal.App("hunyuan_mt")
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake", "ninja-build")
    .pip_install("wheel")
    .pip_install("accelerate", "torch==2.6.0", "transformers==4.56.0")
    # https://github.com/Dao-AILab/flash-attention/releases/tag/v2.7.4.post1
    .pip_install("flash-attn==2.7.4.post1", extra_options="--no-build-isolation")
    .pip_install("compressed-tensors==0.11.0")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "tencent/Hunyuan-MT-7B"),
        }
    )
)

if APP_NAME == "achatbot":
    img = img.pip_install(
        f"achatbot==0.0.24.post2",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)


with img.imports():
    import random

    import torch
    from transformers import AutoModelForCausalLM, AutoTokenizer, AutoConfig
    from transformers.generation.streamers import TextIteratorStreamer

    MODEL_ID = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, MODEL_ID)

    config = None
    if MODEL_ID in ["tencent/Hunyuan-MT-7B-fp8"]:
        # https://huggingface.co/docs/transformers/v4.56.0/en/model_doc/hunyuan_v1_dense#transformers.HunYuanDenseV1Config
        config = AutoConfig.from_pretrained(MODEL_PATH)
        config.quantization_config["ignore"] = config.quantization_config["ignored_layers"]
        del config.quantization_config["ignored_layers"]


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=1,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
async def run(func):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = None
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda")
        print(gpu_prop)

    if asyncio.iscoroutinefunction(func):
        await func(gpu_prop)
    else:
        func(gpu_prop)


def print_model_params(model: torch.nn.Module, extra_info=""):
    # print the number of parameters in the model
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(model)
    print(f"{extra_info} {model_million_params} M parameters")


# https://huggingface.co/collections/tencent/hunyuan-mt-68b42f76d473f82798882597
def dump_model(gpu_prop):
    model = AutoModelForCausalLM.from_pretrained(
        MODEL_PATH,
        device_map="auto",
        dtype="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        config=config,
    )

    model = model.eval()
    # tencent/Hunyuan-MT-7B 7504.932864 M parameters # BF16
    # tencent/Hunyuan-MT-7B-fp8 7504.933312 M parameters ??? # F32 · BF16 · F8_E4M3
    print_model_params(model, f"{MODEL_ID}")
    print(f"{config=}")
    print(f"{model.config=}")
    assert config == model.config

    del model
    torch.cuda.empty_cache()


def tokenize(gpu_prop):
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH, use_fast=True)
    # extra_203, now support 33 languages.
    print(f"{tokenizer=}")
    chats = [
        [
            {
                "role": "user",
                "content": "Translate the following segment into Chinese, without additional explanation.\n\nIt’s on the house.",
            },
        ],
        [
            {
                "role": "user",
                "content": "把下面的文本翻译成English，不要额外解释。\n\n奥利给!",
            },
        ],
    ]
    print(chats)

    # use chat template
    tokens_list = []
    for chat in chats:
        tokens = tokenizer.apply_chat_template(chat, tokenize=False, add_generation_prompt=True)
        tokens_list.append(tokens)

    print(f"{tokens_list=}")

    print("---" * 20)
    # don't use chat template
    inputs = tokenizer(
        tokens_list,
        return_tensors="pt",
        padding=True,  # for batch
    )
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    decoded_prompts = tokenizer.batch_decode(
        input_ids,
        skip_special_tokens=True,
    )
    print(f"{decoded_prompts=}")


def generate(gpu_prop):
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH, use_fast=True)
    # Load model
    model = AutoModelForCausalLM.from_pretrained(
        MODEL_PATH,
        dtype="auto",
        device_map="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        config=config,
    )
    model = model.eval()
    # print(f"{model.config=}")

    messages = [
        "<|startoftext|>Translate the following segment into Chinese, without additional explanation.\n\nIt’s on the house.<|extra_0|>",
        "<|startoftext|>把下面的文本翻译成English，不要额外解释。\n\n奥利给!<|extra_0|>",
    ]
    print(messages)

    inputs = tokenizer(
        messages,
        return_tensors="pt",
        padding=True,  # for batch
        padding_side="left",
    ).to(model.device)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = tokenizer.batch_decode(input_ids)
    print(f"{prompt=}")
    input_len = inputs["input_ids"].shape[-1]

    with torch.inference_mode():
        generated_ids = model.generate(
            input_ids,
            attention_mask=inputs["attention_mask"],
            do_sample=True,
            temperature=0.7,
            top_k=20,
            top_p=0.6,
            repetition_penalty=1.05,
            max_new_tokens=128,
        )
        generated_ids = [generated_id[input_len:] for generated_id in generated_ids]

    for generated_id in generated_ids:
        print(f"{generated_id.shape=}")
    generated_text = tokenizer.batch_decode(
        generated_ids,
        skip_special_tokens=True,
    )
    print(generated_text)


def generate_stream(gpu_prop):
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH, use_fast=True)
    # Load model
    model = AutoModelForCausalLM.from_pretrained(
        MODEL_PATH,
        dtype="auto",
        device_map="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        config=config,
    )
    model = model.eval()
    # print(f"{model.config=}")

    messages = [
        {
            "role": "user",
            "content": "Translate the following segment into Chinese, without additional explanation.\n\nIt’s on the house.",
        },
    ]
    print(messages)

    # inputs = tokenizer.apply_chat_template(
    #    messages,
    #    add_generation_prompt=True,
    #    tokenize=True,
    #    return_dict=True,
    #    return_tensors="pt",
    # ).to(model.device)

    inputs = tokenizer(
        "<|startoftext|>Translate the following segment into Chinese, without additional explanation.\n\nIt’s on the house.<|extra_0|>",
        return_tensors="pt",
    ).to(model.device)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = tokenizer.decode(input_ids[0])
    print(f"{prompt=}")

    streamer = TextIteratorStreamer(tokenizer=tokenizer, skip_prompt=True, skip_special_tokens=True)
    generation_kwargs = dict(
        input_ids=input_ids,
        attention_mask=inputs["attention_mask"],
        do_sample=True,
        temperature=0.7,
        top_k=20,
        top_p=0.6,
        repetition_penalty=1.05,
        max_new_tokens=256,
        streamer=streamer,
    )
    thread = Thread(target=model.generate, kwargs=generation_kwargs)
    thread.start()
    generated_text = ""
    start = perf_counter()
    times = []
    with torch.inference_mode():
        for new_text in streamer:
            times.append(perf_counter() - start)
            print(new_text, end="", flush=True)
            generated_text += new_text
            start = perf_counter()
    print(f"\n{generated_text=} TTFT: {times[0]:.2f}s total time: {sum(times):.2f}s")


async def achatbot_gen_stream(gpu_prop):
    from achatbot.core.llm.transformers.generator import TransformersGenerator, GenerationConfig
    from achatbot.types.llm.transformers import TransformersLMArgs
    from achatbot.common.types import MODELS_DIR, SessionCtx
    from achatbot.common.session import Session
    from achatbot.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    # tokenizer
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH)
    tokens = tokenizer(
        "<|startoftext|>Translate the following segment into Chinese, without additional explanation.\n\nIt’s on the house.<|extra_0|>"
    )
    input_ids = tokens["input_ids"]
    attention_mask = tokens["attention_mask"]

    # init generator
    args = TransformersLMArgs(lm_model_name_or_path=MODEL_PATH).__dict__
    args["config"] = config
    generator = TransformersGenerator(**args)

    # generation_config
    # https://huggingface.co/docs/transformers/v4.56.0/en/main_classes/text_generation#transformers.GenerationConfig
    generation_config = GenerationConfig.from_pretrained(MODEL_PATH, "generation_config.json")
    generation_config.max_new_tokens = 30
    print(generation_config.to_dict())

    session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    session.ctx.state["token_ids"] = input_ids
    first = True
    start_time = time.perf_counter()
    async for token_id in generator.generate(
        session,
        attention_mask=attention_mask,
        stop_ids=[tokenizer.eos_token_id],
        **generation_config.to_dict(),
    ):
        if first:
            ttft = time.perf_counter() - start_time
            print(f"generate TTFT time: {ttft} s")
            first = False
        gen_text = tokenizer.decode(token_id)
        print(token_id, gen_text)


# !!! If you want to load fp8 model with transformers, you need to change the name"ignored_layers" in config.json to "ignore" and upgrade the compressed-tensors to compressed-tensors-0.11.0.

"""
modal run src/llm/transformers/translation/hunyuan_mt.py --task tokenize

IMAGE_GPU=T4 modal run src/llm/transformers/translation/hunyuan_mt.py --task dump_model
LLM_MODEL=tencent/Hunyuan-MT-7B-fp8 IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task dump_model

IMAGE_GPU=T4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate
IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate
LLM_MODEL=tencent/Hunyuan-MT-7B-fp8 IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate

IMAGE_GPU=T4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate_stream
IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate_stream
LLM_MODEL=tencent/Hunyuan-MT-7B-fp8 IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task generate_stream

APP_NAME=achatbot IMAGE_GPU=T4 modal run src/llm/transformers/translation/hunyuan_mt.py --task achatbot_gen_stream
APP_NAME=achatbot IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task achatbot_gen_stream
LLM_MODEL=tencent/Hunyuan-MT-7B-fp8 APP_NAME=achatbot IMAGE_GPU=L4 modal run src/llm/transformers/translation/hunyuan_mt.py --task achatbot_gen_stream
"""


@app.local_entrypoint()
def main(task: str = "dump_model"):
    print(task)
    tasks = {
        "tokenize": tokenize,
        "dump_model": dump_model,
        "generate": generate,
        "generate_stream": generate_stream,
        "achatbot_gen_stream": achatbot_gen_stream,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



import soundfile as sf
import torch
import fire
import torch.nn.functional as F
from torchaudio.functional import resample
from models.ecapa_tdnn import ECAPA_TDNN_SMALL

MODEL_LIST = ['ecapa_tdnn', 'hubert_large', 'wav2vec2_xlsr', 'unispeech_sat']


def init_model(model_name, checkpoint=None):
    if model_name == 'unispeech_sat':
        config_path = 'config/unispeech_sat.th'
        model = ECAPA_TDNN_SMALL(feat_dim=1024, feat_type='unispeech_sat', config_path=config_path)
    elif model_name == 'hubert_large':
        config_path = None
        model = ECAPA_TDNN_SMALL(feat_dim=1024, feat_type='hubert_large_ll60k', config_path=config_path)
    elif model_name == 'wav2vec2_xlsr':
        config_path = None
        model = ECAPA_TDNN_SMALL(feat_dim=1024, feat_type='wav2vec2_xlsr', config_path=config_path)
    else:
        model = ECAPA_TDNN_SMALL(feat_dim=40, feat_type='fbank')

    if checkpoint is not None:
        state_dict = torch.load(checkpoint, map_location=lambda storage, loc: storage)
        model.load_state_dict(state_dict['model'], strict=False)
    return model


def verification(model_name,  wav1, wav2, use_gpu=True, checkpoint=None):

    assert model_name in MODEL_LIST, 'The model_name should be in {}'.format(MODEL_LIST)
    model = init_model(model_name, checkpoint)

    wav1, sr1 = sf.read(wav1)
    wav2, sr2 = sf.read(wav2)

    wav1 = torch.from_numpy(wav1).unsqueeze(0).float()
    wav2 = torch.from_numpy(wav2).unsqueeze(0).float()
    wav1 = resample(wav1, orig_freq=sr1, new_freq=16000)
    wav2 = resample(wav2, orig_freq=sr2, new_freq=16000)

    if use_gpu:
        model = model.cuda()
        wav1 = wav1.cuda()
        wav2 = wav2.cuda()

    model.eval()
    with torch.no_grad():
        emb1 = model(wav1)
        emb2 = model(wav2)

    sim = F.cosine_similarity(emb1, emb2)
    print("The similarity score between two audios is {:.4f} (-1.0, 1.0).".format(sim[0].item()))


if __name__ == "__main__":
    fire.Fire(verification)




import logging
from typing import Optional
import uuid

from apipeline.processors.async_frame_processor import AsyncFrameProcessor


class MemoryProcessor(AsyncFrameProcessor):
    def __init__(
        self,
        user_id: str | None = None,
        agent_id: str | None = None,
        run_id: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.user_id = user_id or f"user_{str(uuid.uuid4())}"
        self.agent_id = agent_id or f"agent_{str(uuid.uuid4())}"
        self.run_id = run_id or f"run_{str(uuid.uuid4())}"

    def set_user_id(self, user_id):
        self.user_id = user_id

    def set_agent_id(self, agent_id):
        self.agent_id = agent_id

    def set_run_id(self, run_id):
        self.run_id = run_id

    async def get_all_memories(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        **kwargs,
    ):
        raise NotImplementedError("Subclasses must implement get_all_memories method")

    async def get_initial_memories(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Fetch all memories for the user to create a personalized greeting with memories.

        Returns:
            A personalized greeting based on user memories
        """
        try:
            memories = await self.get_all_memories(
                user_id=user_id, agent_id=agent_id, run_id=run_id, **kwargs
            )

            if not memories or not memories.get("results"):
                logging.debug(
                    f"!!! No memories found for this user {user_id}, agent {agent_id}, run {run_id}"
                )
                return ""

            # Create a personalized greeting based on memories
            memory_str = ""
            # Add some personalization based on memories (limit to 3 memories for brevity)
            for i, memory in enumerate(memories["results"][:3]):
                memory_content = memory.get("memory", "")
                # Keep memory references brief
                if len(memory_content) > 100:
                    memory_content = memory_content[:97] + "..."
                memory_str += f"{memory_content} "

            logging.debug(
                f"get {len(memories)} memories for this user {user_id}, agent {agent_id}, run {run_id}"
            )
            return memory_str

        except Exception as e:
            logging.error(f"Error retrieving initial memories from Mem0: {e}")
            return ""



import subprocess

import modal

model_dir = "/root/models"
MODEL_VOL = modal.Volume.from_name("models", create_if_missing=True)

# completed workflows write output images/video/audio to this directory
comfyui_out_dir = "/root/comfy/ComfyUI/output"
COMFYUI_OUT_VOL = modal.Volume.from_name("comfyui_output", create_if_missing=True)


# https://docs.comfy.org/installation/manual_install#example-structure
def clear_comfyui_output_dir():
    # clear output file
    subprocess.run(
        f"rm -f {comfyui_out_dir}/_output_images_will_be_put_here", shell=True, check=True
    )



# Copyright ESPnet (https://github.com/espnet/espnet). All Rights Reserved.
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)

from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Tuple

import torch


class AbsESPnetModel(torch.nn.Module, ABC):
    """The common abstract class among each tasks

    "ESPnetModel" is referred to a class which inherits torch.nn.Module,
    and makes the dnn-models forward as its member field,
    a.k.a delegate pattern,
    and defines "loss", "stats", and "weight" for the task.

    If you intend to implement new task in ESPNet,
    the model must inherit this class.
    In other words, the "mediator" objects between
    our training system and the your task class are
    just only these three values, loss, stats, and weight.

    Example:
        >>> from funasr_local.tasks.abs_task import AbsTask
        >>> class YourESPnetModel(AbsESPnetModel):
        ...     def forward(self, input, input_lengths):
        ...         ...
        ...         return loss, stats, weight
        >>> class YourTask(AbsTask):
        ...     @classmethod
        ...     def build_model(cls, args: argparse.Namespace) -> YourESPnetModel:
    """

    def __init__(self):
        super().__init__()
        self.num_updates = 0

    @abstractmethod
    def forward(
        self, **batch: torch.Tensor
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor], torch.Tensor]:
        raise NotImplementedError

    @abstractmethod
    def collect_feats(self, **batch: torch.Tensor) -> Dict[str, torch.Tensor]:
        raise NotImplementedError

    def set_num_updates(self, num_updates):
        self.num_updates = num_updates

    def get_num_updates(self):
        return self.num_updates



import os

from funasr import AutoModel
import librosa
import soundfile
import typer

app = typer.Typer()


@app.command()
def online():
    """
    KPI:...
    - https://www.modelscope.cn/models/iic/speech_paraformer_asr_nat-zh-cn-16k-common-vocab8404-online/summary
    - https://www.modelscope.cn/models/iic/speech_paraformer-large_asr_nat-zh-cn-16k-common-vocab8404-online/summary
    """
    chunk_size = [0, 10, 5]  # [0, 10, 5] 600ms
    # chunk_size = [0, 8, 4]  # [0, 8, 4] 480ms
    encoder_chunk_look_back = 4  # number of chunks to lookback for encoder self-attention
    decoder_chunk_look_back = 1  # number of encoder chunks to lookback for decoder cross-attention

    model = AutoModel(
        # model="iic/speech_paraformer_asr_nat-zh-cn-16k-common-vocab8404-online",
        # model="iic/speech_paraformer-large_asr_nat-zh-cn-16k-common-vocab8404-online",
        model="paraformer-zh-streaming",
        model_revision="v2.0.4",
    )  # large
    with open("./paraformer_model.txt", "w") as f:
        print(model.model, file=f, flush=True)

    # wav_file = os.path.join(model.model_path, "example/asr_example.wav")
    wav_file = "./assets/Chinese_prompt.wav"
    print(wav_file)
    speech, sample_rate = soundfile.read(wav_file)
    print(sample_rate, speech.shape)
    target_sample_rate = 16000
    if sample_rate != target_sample_rate:
        speech = librosa.resample(speech, orig_sr=sample_rate, target_sr=target_sample_rate)
        sample_rate = target_sample_rate
        print(f"Resampled to {target_sample_rate}Hz {speech.shape=}")

    chunk_stride = chunk_size[1] * 960  # 600ms

    cache = {}
    total_chunk_num = int(len((speech) - 1) / chunk_stride + 1)
    for i in range(total_chunk_num):
        speech_chunk = speech[i * chunk_stride : (i + 1) * chunk_stride]
        is_final = i == total_chunk_num - 1
        res = model.generate(
            input=speech_chunk,
            cache=cache,
            is_final=is_final,
            chunk_size=chunk_size,
            encoder_chunk_look_back=encoder_chunk_look_back,
            decoder_chunk_look_back=decoder_chunk_look_back,
            disable_pbar=True,
        )
        print(res)


@app.command()
def offline():
    """
    - KPI: ...
    - https://www.modelscope.cn/models/iic/speech_paraformer-large_asr_nat-zh-cn-16k-common-vocab8404-pytorch
    - long https://www.modelscope.cn/models/iic/speech_paraformer-large-vad-punc_asr_nat-zh-cn-16k-common-vocab8404-pytorch
    - speaker https://modelscope.cn/models/damo/speech_paraformer-large-vad-punc-spk_asr_nat-zh-cn
    - hotword https://www.modelscope.cn/models/iic/speech_paraformer-large-contextual_asr_nat-zh-cn-16k-common-vocab8404
    """
    # paraformer-zh is a multi-functional asr model
    # use vad, punc, spk or not as you need
    model = AutoModel(
        model="paraformer-zh",
        model_revision="v2.0.4",
        vad_model="fsmn-vad",
        vad_model_revision="v2.0.4",
        punc_model="ct-punc-c",  # https://modelscope.cn/models/iic/punc_ct-transformer_zh-cn-common-vad_realtime-vocab272727/summary
        punc_model_revision="v2.0.4",
        # spk_model="cam++", spk_model_revision="v2.0.2",
    )
    with open("./paraformer_model.txt", "w") as f:
        print(model.model, file=f, flush=True)
    input = "./assets/Chinese_prompt.wav"
    # input = f"{model.model_path}/example/asr_example.wav"
    res = model.generate(input=input, batch_size_s=300, hotword="魔搭", disable_pbar=True)
    print(res)


"""
python demo/modelscope/paraformer_asr.py offline
python demo/modelscope/paraformer_asr.py online
"""
if __name__ == "__main__":
    app()

"""
Cif:
  (predictor): CifPredictorV2(
    (pad): ConstantPad1d(padding=(1, 1), value=0)
    (cif_conv1d): Conv1d(320, 320, kernel_size=(3,), stride=(1,))
    (cif_output): Linear(in_features=320, out_features=1, bias=True)
    (dropout): Dropout(p=0.1, inplace=False)
  )
"""



import logging
import unittest
import os

import wave
from pydub import AudioSegment

from src.common.logger import Logger
from src.common.session import Session
from src.common.interface import EndOfTurnState, ITurnAnalyzer
from src.common.types import SessionCtx, TEST_DIR, MODELS_DIR, RECORDS_DIR
from src.modules.speech.turn_analyzer import TurnAnalyzerEnvInit


r"""
LOG_LEVEL=debug TURN_ANALYZER_TAG=v2_smart_turn_analyzer python -m unittest test.modules.speech.turn_analyzer.test_smart_turn.TestTurnAnalyzer.test_end
TURN_TORCH_DTYPE=bfloat16 TURN_ANALYZER_TAG=v2_smart_turn_analyzer python -m unittest test.modules.speech.turn_analyzer.test_smart_turn.TestTurnAnalyzer.test_end

TURN_ANALYZER_TAG=v2_smart_turn_analyzer python -m unittest test.modules.speech.turn_analyzer.test_smart_turn.TestTurnAnalyzer.test_no_end
"""


class TestTurnAnalyzer(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        # wget
        # https://isv-data.oss-cn-hangzhou.aliyuncs.com/ics/MaaS/ASR/test_audio/asr_example_zh.wav
        # -O records/asr_example_zh.wav
        cls.audio_file = os.path.join(TEST_DIR, "audio_files/asr_example_zh.wav")
        # Use an environment variable to get the ASR model TAG
        cls.tag = os.getenv("TURN_ANALYZER_TAG", "v2_smart_turn_analyzer")
        cls.model_name_or_path = os.getenv(
            "TURN_MODEL_PATH", os.path.join(MODELS_DIR, "pipecat-ai/smart-turn-v2")
        )
        cls.torch_dtype = os.getenv("TURN_TORCH_DTYPE", "float32")
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        kwargs = {}
        kwargs["model_path"] = self.model_name_or_path
        kwargs["torch_dtype"] = self.torch_dtype

        self.turn: ITurnAnalyzer = TurnAnalyzerEnvInit.initTurnAnalyzerEngine(self.tag, kwargs)

        self.session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)

    async def asyncTearDown(self):
        pass

    async def test_end(self):
        assert self.turn.speech_triggered is False

        audio_bytes = b""
        with open(self.audio_file, "rb") as audio_file:
            audio_bytes = audio_file.read()
        state = self.turn.append_audio(audio_bytes, True)
        # print(state)
        assert self.turn.speech_triggered is True
        assert state == EndOfTurnState.INCOMPLETE

        state, result = await self.turn.analyze_end_of_turn()
        # print(state, result)
        assert self.turn.speech_triggered is False
        assert state == EndOfTurnState.COMPLETE
        assert result["prediction"] == 1

        self.turn.clear()
        assert self.turn.speech_triggered is False

    async def test_no_end(self):
        assert self.turn.speech_triggered is False

        audio = AudioSegment.from_file(self.audio_file, format="wav")

        total_duration = len(audio)
        duration_to_capture = int(total_duration * 3 / 4)
        audio_clip = audio[:duration_to_capture]

        tmp_path = os.path.join(RECORDS_DIR, "turn_temp_clip.wav")
        audio_clip.export(tmp_path, format="wav")
        print(f"Exported temp clip to {tmp_path}")
        assert os.path.exists(tmp_path)

        with wave.open(tmp_path, "rb") as wav_file:
            n_channels = wav_file.getnchannels()
            sample_width = wav_file.getsampwidth()
            frame_rate = wav_file.getframerate()
            n_frames = wav_file.getnframes()
            audio_bytes = wav_file.readframes(n_frames)

            print(f"Channels: {n_channels}")
            print(f"Sample Width: {sample_width}")
            print(f"Frame Rate: {frame_rate}")
            print(f"Number of Frames: {n_frames}")
            print(f"Audio Data length: {len(audio_bytes)}")

            state = self.turn.append_audio(audio_bytes, True)
            # print(state)
            assert self.turn.speech_triggered is True
            assert state == EndOfTurnState.INCOMPLETE

            state, result = await self.turn.analyze_end_of_turn()
            print(state, result)
            assert self.turn.speech_triggered is True
            assert state == EndOfTurnState.INCOMPLETE
            assert result["prediction"] != 1

            self.turn.clear()
            assert self.turn.speech_triggered is False



import glob
import io
import logging
import math
import os
import tarfile
import uuid

import safetensors
import torch
from transformers import WhisperFeatureExtractor, WhisperTokenizerFast

import torchaudio

from transformers import WhisperFeatureExtractor
from speech_tokenizer.modeling_whisper import WhisperVQEncoder
from flow_inference import AudioDecoder

from funasr.utils.load_utils import load_audio_text_image_video, extract_fbank
from funasr.models.sense_voice.model import SenseVoiceSmall

from .constants import (
    AUD_CONTEXT_TOKEN,
    AUD_END_TOKEN,
    AUD_START_TOKEN,
    AUD_TAG_TOKEN,
    BOX_END_TOKEN,
    BOX_START_TOKEN,
    IMG_CONTEXT_TOKEN,
    IMG_END_TOKEN,
    IMG_START_TOKEN,
    IMG_TAG_TOKEN,
    PATCH_CONTEXT_TOKEN,
    PATCH_END_TOKEN,
    PATCH_START_TOKEN,
    QUAD_END_TOKEN,
    QUAD_START_TOKEN,
    REF_END_TOKEN,
    REF_START_TOKEN,
    VID_CONTEXT_TOKEN,
    VID_END_TOKEN,
    VID_START_TOKEN,
    VID_TAG_TOKEN,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def update_tokenizer_for_sensevoice_glm4voice(tokenizer):
    token_list = [
        IMG_START_TOKEN,
        IMG_END_TOKEN,
        IMG_CONTEXT_TOKEN,
        VID_START_TOKEN,
        VID_END_TOKEN,
        VID_CONTEXT_TOKEN,
        PATCH_START_TOKEN,
        PATCH_END_TOKEN,
        PATCH_CONTEXT_TOKEN,
        AUD_START_TOKEN,
        AUD_END_TOKEN,
        AUD_CONTEXT_TOKEN,
        QUAD_START_TOKEN,
        QUAD_END_TOKEN,
        REF_START_TOKEN,
        REF_END_TOKEN,
        BOX_START_TOKEN,
        BOX_END_TOKEN,
        IMG_TAG_TOKEN,
        VID_TAG_TOKEN,
        AUD_TAG_TOKEN,
    ]
    num_new_tokens = tokenizer.add_tokens(token_list, special_tokens=True)

    token_list = [f"<|audio_{i}|>" for i in range(16384)]
    num_new_tokens = tokenizer.add_tokens(token_list, special_tokens=False)

    # logger.info(f"tokenizer {tokenizer}")
    return tokenizer


class SenseVoiceGLM4VoiceTokenizer:
    def __init__(
        self,
        glm4_voice_tokenizer_model_path=None,
        sense_voice_model_path=None,
        flow_path=None,
        rank=None,
    ):
        self.sense_voice_model_path = sense_voice_model_path
        self.glm4_voice_tokenizer_model_path = glm4_voice_tokenizer_model_path
        self.flow_path = flow_path

        if rank is None and torch.distributed.is_initialized():
            rank = torch.distributed.get_rank()
            self.rank = rank % 8
        else:
            self.rank = rank
        logger.info(f"{self.rank=}")

        self.sample_rate = 16000

        self.is_discrete = True
        self.is_contiguous = True

        # #                            T   A
        # text_audio_interval_ratio = [13, 26]
        # #                            T  A  T  A  T  A
        # text_audio_interval_ratio = [1, 4, 3, 8, 4, 10]
        # #                            T  A   T  A
        # text_audio_interval_ratio = [1, 10, 4, 10]

        # self.text_audio_interval_ratio = text_audio_interval_ratio

    def load_model(self):
        import faulthandler
        faulthandler.enable()

        if self.rank is not None:
            self.device = f"cuda:{self.rank}"
            torch.cuda.set_device(self.rank)
        else:
            self.device = "cpu"

        if self.sense_voice_model_path is not None:
            logger.info(f"{self.device=} Loading SenseVoiceSmall")
            self.sensevoice_model, self.kwargs = SenseVoiceSmall.from_pretrained(
                model=self.sense_voice_model_path, device=self.device)
            logger.info(f"{self.device=} Loading SenseVoiceSmall Done")

        if self.glm4_voice_tokenizer_model_path is not None:
            logger.info(f"{self.device=} Loading GLM4VoiceTokenizer")
            self.whisper_model = (
                WhisperVQEncoder.from_pretrained(
                    self.glm4_voice_tokenizer_model_path).eval().to(
                    self.device))
            self.feature_extractor = WhisperFeatureExtractor.from_pretrained(
                self.glm4_voice_tokenizer_model_path)
            logger.info(f"{self.device=} Loading GLM4VoiceTokenizer Done")

        if self.flow_path is not None:
            logger.info(f"{self.device=} Loading GLM4VoiceDecoder")
            flow_config = os.path.join(self.flow_path, "config.yaml")
            flow_checkpoint = os.path.join(self.flow_path, "flow.pt")
            hift_checkpoint = os.path.join(self.flow_path, "hift.pt")

            # Flow & Hift
            self.audio_decoder = AudioDecoder(
                config_path=flow_config,
                flow_ckpt_path=flow_checkpoint,
                hift_ckpt_path=hift_checkpoint,
                device=self.device,
            )
            logger.info(f"{self.device=} Loading GLM4VoiceDecoder Done")

    def encode(self, utt, is_discrete=False, is_contiguous=True, **kwargs):
        assert not (is_discrete and is_contiguous)
        assert is_discrete or is_contiguous

        if is_discrete:
            audio_tokens = extract_speech_token(
                self.whisper_model, self.feature_extractor, [utt], device=self.device
            )[0]
            return audio_tokens

        if is_contiguous:
            if isinstance(utt, tuple):
                audio, sample_rate = utt
            else:
                audio, sample_rate = torchaudio.load(utt)
            audio = audio.mean(0)
            if sample_rate != self.sample_rate:
                if sample_rate not in _resample_buffer:
                    _resample_buffer[sample_rate] = torchaudio.transforms.Resample(
                        orig_freq=sample_rate, new_freq=self.sample_rate
                    ).to(self.device)
                audio = audio.to(self.device)
                audio = _resample_buffer[sample_rate](audio[None, :])[0, :]
                audio = audio.cpu()
            # resampler = torchaudio.transforms.Resample(
            #     orig_freq=sample_rate, new_freq=self.sample_rate
            # )
            # audio = resampler(audio[None, :])[0, :]
            # audio = audio.to(self.device)

            frontend = self.kwargs["frontend"]

            speech, speech_lengths = extract_fbank(audio, data_type="sound", frontend=frontend)

            speech = speech[0]
            # print(f"{speech_lengths=}")
            # print(f"{speech.size()=}")

            return speech

    def decode(self, audio_tokens, option_steps=10, **kwargs):
        this_uuid = kwargs.get("session_id","abc") or "abc"

        tts_token = torch.tensor(audio_tokens, device=self.device).unsqueeze(0)

        flow_prompt_speech_token = torch.zeros(1, 0, dtype=torch.int64).to(self.device)
        prompt_speech_feat = torch.zeros(1, 0, 80).to(self.device)

        tts_speech, tts_mel = self.audio_decoder.token2wav(
            tts_token,
            uuid=this_uuid,
            prompt_token=flow_prompt_speech_token.to(self.device),
            prompt_feat=prompt_speech_feat.to(self.device),
            finalize=True,
            option_steps=option_steps,
        )
        tts_speechs = []
        tts_speechs.append(tts_speech.squeeze())
        tts_speech = torch.cat(tts_speechs, dim=-1).cpu()

        return tts_speech

    def apply_to_role(self, role, **kwargs):
        is_discrete = kwargs.get("is_discrete", False)
        if is_discrete and role in ["assistant", "gpt"]:
            return True

        is_contiguous = kwargs.get("is_contiguous", False)
        if is_contiguous and role in ["user", "human"]:
            return True

        return False


_resample_buffer: dict[int, torchaudio.transforms.Resample] = {}


def extract_speech_token(model, feature_extractor, utts, device="cuda"):
    with torch.no_grad():
        audios, indices = [], []
        for idx, utt in enumerate(utts):
            if isinstance(utt, tuple):
                audio, sample_rate = utt
            else:
                audio, sample_rate = torchaudio.load(utt)
            audio = audio.to(device)
            if sample_rate != 16000:
                if sample_rate not in _resample_buffer:
                    _resample_buffer[sample_rate] = torchaudio.transforms.Resample(
                        orig_freq=sample_rate, new_freq=16000
                    ).to(device)
                audio = _resample_buffer[sample_rate](audio)
            # if audio.shape[0] > 1:
            #     audio = audio[:1]
            audio = audio[0]
            audio = audio.cpu().numpy()
            time_step = 0
            while time_step * 16000 < audio.shape[0]:
                audio_segment = audio[time_step * 16000: (time_step + 30) * 16000]
                audios.append(audio_segment)
                indices.append(idx)
                time_step += 30
        pooling_kernel_size = model.config.pooling_kernel_size or 1
        stride = (
            model.conv1.stride[0]
            * model.conv2.stride[0]
            * pooling_kernel_size
            * feature_extractor.hop_length
        )
        all_speech_tokens = [[] for _ in range(len(utts))]
        batch_size = 128
        for start in range(0, len(audios), batch_size):
            features = feature_extractor(
                audios[start: start + batch_size],
                sampling_rate=16000,
                return_attention_mask=True,
                return_tensors="pt",
                device=device,
                padding="longest",
                pad_to_multiple_of=stride,
            )
            features = features.to(device=device)
            outputs = model(**features)
            speech_tokens = outputs.quantized_token_ids
            attention_mask = features.attention_mask[
                :, :: model.conv1.stride[0] * model.conv2.stride[0]
            ]
            attention_mask = attention_mask[:, :: model.config.pooling_kernel_size]
            assert attention_mask.shape == speech_tokens.shape
            for i in range(len(speech_tokens)):
                idx = indices[start + i]
                speech_token = speech_tokens[i][attention_mask[i].bool()].tolist()
                all_speech_tokens[idx].extend(speech_token)
        return all_speech_tokens



from typing import List, Tuple

import torch
import torchaudio
from torch import nn

from decoder.modules import safe_log

import torch.nn.functional as F


class MelSpecReconstructionLoss(nn.Module):
    """
    L1 distance between the mel-scaled magnitude spectrograms of the ground truth sample and the generated sample
    """

    def __init__(
        self, sample_rate: int = 24000, n_fft: int = 1024, hop_length: int = 256, n_mels: int = 100,
    ):
        super().__init__()
        self.mel_spec = torchaudio.transforms.MelSpectrogram(
            sample_rate=sample_rate, n_fft=n_fft, hop_length=hop_length, n_mels=n_mels, center=True, power=1,
        )

    def forward(self, y_hat, y) -> torch.Tensor:
        """
        Args:
            y_hat (Tensor): Predicted audio waveform.
            y (Tensor): Ground truth audio waveform.

        Returns:
            Tensor: L1 loss between the mel-scaled magnitude spectrograms.
        """
        mel_hat = safe_log(self.mel_spec(y_hat))
        mel = safe_log(self.mel_spec(y))

        loss = torch.nn.functional.l1_loss(mel, mel_hat)

        return loss


class GeneratorLoss(nn.Module):
    """
    Generator Loss module. Calculates the loss for the generator based on discriminator outputs.
    """

    def forward(self, disc_outputs: List[torch.Tensor]) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Args:
            disc_outputs (List[Tensor]): List of discriminator outputs.

        Returns:
            Tuple[Tensor, List[Tensor]]: Tuple containing the total loss and a list of loss values from
                                         the sub-discriminators
        """
        loss = 0
        gen_losses = []
        for dg in disc_outputs:
            l = torch.mean(torch.clamp(1 - dg, min=0))
            gen_losses.append(l)
            loss += l

        return loss, gen_losses


class DiscriminatorLoss(nn.Module):
    """
    Discriminator Loss module. Calculates the loss for the discriminator based on real and generated outputs.
    """

    def forward(
        self, disc_real_outputs: List[torch.Tensor], disc_generated_outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor], List[torch.Tensor]]:
        """
        Args:
            disc_real_outputs (List[Tensor]): List of discriminator outputs for real samples.
            disc_generated_outputs (List[Tensor]): List of discriminator outputs for generated samples.

        Returns:
            Tuple[Tensor, List[Tensor], List[Tensor]]: A tuple containing the total loss, a list of loss values from
                                                       the sub-discriminators for real outputs, and a list of
                                                       loss values for generated outputs.
        """
        loss = 0
        r_losses = []
        g_losses = []
        for dr, dg in zip(disc_real_outputs, disc_generated_outputs):
            r_loss = torch.mean(torch.clamp(1 - dr, min=0))
            g_loss = torch.mean(torch.clamp(1 + dg, min=0))
            loss += r_loss + g_loss
            r_losses.append(r_loss.item())
            g_losses.append(g_loss.item())

        return loss, r_losses, g_losses


class FeatureMatchingLoss(nn.Module):
    """
    Feature Matching Loss module. Calculates the feature matching loss between feature maps of the sub-discriminators.
    """

    def forward(self, fmap_r: List[List[torch.Tensor]], fmap_g: List[List[torch.Tensor]]) -> torch.Tensor:
        """
        Args:
            fmap_r (List[List[Tensor]]): List of feature maps from real samples.
            fmap_g (List[List[Tensor]]): List of feature maps from generated samples.

        Returns:
            Tensor: The calculated feature matching loss.
        """
        loss = 0
        for dr, dg in zip(fmap_r, fmap_g):
            for rl, gl in zip(dr, dg):
                loss += torch.mean(torch.abs(rl - gl))

        return loss

class DACGANLoss(nn.Module):
    """
    Computes a discriminator loss, given a discriminator on
    generated waveforms/spectrograms compared to ground truth
    waveforms/spectrograms. Computes the loss for both the
    discriminator and the generator in separate functions.
    """

    def __init__(self, discriminator):
        super().__init__()
        self.discriminator = discriminator

    def forward(self, fake, real):
        # d_fake = self.discriminator(fake.audio_data)
        # d_real = self.discriminator(real.audio_data)
        d_fake = self.discriminator(fake)
        d_real = self.discriminator(real)
        return d_fake, d_real

    def discriminator_loss(self, fake, real):
        d_fake, d_real = self.forward(fake.clone().detach(), real)

        loss_d = 0
        for x_fake, x_real in zip(d_fake, d_real):
            loss_d += torch.mean(x_fake[-1] ** 2)
            loss_d += torch.mean((1 - x_real[-1]) ** 2)
        return loss_d

    def generator_loss(self, fake, real):
        d_fake, d_real = self.forward(fake, real)

        loss_g = 0
        for x_fake in d_fake:
            loss_g += torch.mean((1 - x_fake[-1]) ** 2)

        loss_feature = 0

        for i in range(len(d_fake)):
            for j in range(len(d_fake[i]) - 1):
                loss_feature += F.l1_loss(d_fake[i][j], d_real[i][j].detach())
        return loss_g, loss_feature




""" from https://github.com/keithito/tacotron

Cleaners are transformations that run over the input text at both training and eval time.

Cleaners can be selected by passing a comma-delimited list of cleaner names as the "cleaners"
hyperparameter. Some cleaners are English-specific. You'll typically want to use:
  1. "english_cleaners" for English text
  2. "transliteration_cleaners" for non-English text that can be transliterated to ASCII using
     the Unidecode library (https://pypi.python.org/pypi/Unidecode)
  3. "basic_cleaners" if you do not want to transliterate (in this case, you should also update
     the symbols in symbols.py to match your data).
"""

import logging
import re

import phonemizer
import piper_phonemize
from unidecode import unidecode

# To avoid excessive logging we set the log level of the phonemizer package to Critical
critical_logger = logging.getLogger("phonemizer")
critical_logger.setLevel(logging.CRITICAL)

# Intializing the phonemizer globally significantly reduces the speed
# now the phonemizer is not initialising at every call
# Might be less flexible, but it is much-much faster
global_phonemizer = phonemizer.backend.EspeakBackend(
    language="en-us",
    preserve_punctuation=True,
    with_stress=True,
    language_switch="remove-flags",
    logger=critical_logger,
)


# Regular expression matching whitespace:
_whitespace_re = re.compile(r"\s+")

# List of (regular expression, replacement) pairs for abbreviations:
_abbreviations = [
    (re.compile("\\b%s\\." % x[0], re.IGNORECASE), x[1])
    for x in [
        ("mrs", "misess"),
        ("mr", "mister"),
        ("dr", "doctor"),
        ("st", "saint"),
        ("co", "company"),
        ("jr", "junior"),
        ("maj", "major"),
        ("gen", "general"),
        ("drs", "doctors"),
        ("rev", "reverend"),
        ("lt", "lieutenant"),
        ("hon", "honorable"),
        ("sgt", "sergeant"),
        ("capt", "captain"),
        ("esq", "esquire"),
        ("ltd", "limited"),
        ("col", "colonel"),
        ("ft", "fort"),
    ]
]


def expand_abbreviations(text):
    for regex, replacement in _abbreviations:
        text = re.sub(regex, replacement, text)
    return text


def lowercase(text):
    return text.lower()


def collapse_whitespace(text):
    return re.sub(_whitespace_re, " ", text)


def convert_to_ascii(text):
    return unidecode(text)


def basic_cleaners(text):
    """Basic pipeline that lowercases and collapses whitespace without transliteration."""
    text = lowercase(text)
    text = collapse_whitespace(text)
    return text


def transliteration_cleaners(text):
    """Pipeline for non-English text that transliterates to ASCII."""
    text = convert_to_ascii(text)
    text = lowercase(text)
    text = collapse_whitespace(text)
    return text


def english_cleaners2(text):
    """Pipeline for English text, including abbreviation expansion. + punctuation + stress"""
    text = convert_to_ascii(text)
    text = lowercase(text)
    text = expand_abbreviations(text)
    phonemes = global_phonemizer.phonemize([text], strip=True, njobs=1)[0]
    phonemes = collapse_whitespace(phonemes)
    return phonemes


def english_cleaners_piper(text):
    """Pipeline for English text, including abbreviation expansion. + punctuation + stress"""
    text = convert_to_ascii(text)
    text = lowercase(text)
    text = expand_abbreviations(text)
    phonemes = "".join(piper_phonemize.phonemize_espeak(text=text, voice="en-US")[0])
    phonemes = collapse_whitespace(phonemes)
    return phonemes



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
LayerDrop as described in https://arxiv.org/abs/1909.11556.
"""

import torch
import torch.nn as nn


class LayerDropModuleList(nn.ModuleList):
    """
    A LayerDrop implementation based on :class:`torch.nn.ModuleList`.

    We refresh the choice of which layers to drop every time we iterate
    over the LayerDropModuleList instance. During evaluation we always
    iterate over all layers.

    Usage::

        layers = LayerDropList(p=0.5, modules=[layer1, layer2, layer3])
        for layer in layers:  # this might iterate over layers 1 and 3
            x = layer(x)
        for layer in layers:  # this might iterate over all layers
            x = layer(x)
        for layer in layers:  # this might not iterate over any layers
            x = layer(x)

    Args:
        p (float): probability of dropping out each layer
        modules (iterable, optional): an iterable of modules to add
    """

    def __init__(self, p, modules=None):
        super().__init__(modules)
        self.p = p

    def __iter__(self):
        dropout_probs = torch.empty(len(self)).uniform_()
        for i, m in enumerate(super().__iter__()):
            if not self.training or (dropout_probs[i] > self.p):
                yield m



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from fairseq.data.encoders import register_bpe


SPACE = chr(32)
SPACE_ESCAPE = chr(9601)


@register_bpe("characters")
class Characters(object):
    def __init__(self, *unused):
        pass

    @staticmethod
    def add_args(parser):
        pass

    @staticmethod
    def encode(x: str) -> str:
        escaped = x.replace(SPACE, SPACE_ESCAPE)
        return SPACE.join(list(escaped))

    @staticmethod
    def decode(x: str) -> str:
        return x.replace(SPACE, "").replace(SPACE_ESCAPE, SPACE)



import logging
import time
import uuid
import aiofiles

from dotenv import load_dotenv
from fastapi import WebSocket
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger
from apipeline.frames import AudioRawFrame, TextFrame, Frame, CancelFrame, EndFrame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor

from achatbot.common.session import Session, SessionCtx
from achatbot.serializers.transcription_protobuf import TranscriptionFrameSerializer
from achatbot.processors.speech.vad_audio_save_processor import VADAudioSaveProcessor
from achatbot.modules.speech.asr_live import ASRLiveEnvInit
from achatbot.cmd.bots.base_fastapi_websocket_server import AIFastapiWebsocketBot
from achatbot.modules.speech.vad_analyzer import VADAnalyzerEnvInit
from achatbot.cmd.bots import register_ai_fastapi_ws_bots
from achatbot.types.network.fastapi_websocket import FastapiWebsocketServerParams
from achatbot.transports.fastapi_websocket_server import FastapiWebsocketTransport
from achatbot.processors.speech.asr.asr_live_processor import ASRLiveProcessor
from achatbot.processors.punctuation_processor import PunctuationProcessor
from achatbot.modules.punctuation import PuncEnvInit

load_dotenv(override=True)


class SaveASRText(FrameProcessor):
    def __init__(self, *, is_save=True, name=None, loop=None, **kwargs):
        super().__init__(name=name, loop=loop, **kwargs)
        cur_time = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        self.save_file = f"asr_output_{cur_time}.txt"
        self.asr_text = ""
        self.is_save = is_save

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)
        await self.push_frame(frame, direction)
        if isinstance(frame, TextFrame):
            self.asr_text += frame.text
        if isinstance(frame, (CancelFrame, EndFrame)):
            if not self.is_save:
                return
            # save asr text to file
            async with aiofiles.open(self.save_file, "w", encoding="utf-8") as f:
                await f.write(self.asr_text)


@register_ai_fastapi_ws_bots.register
class FastapiWebsocketStreamingASRBot(AIFastapiWebsocketBot):
    """
    fastapi websocket input(audio)/output(text) server bot with vad,asr,punc,nt
    """

    def __init__(self, websocket: WebSocket | None = None, **args) -> None:
        super().__init__(websocket=websocket, **args)
        self.init_bot_config()

        self.vad_analyzer = None
        self.asr_engine = None
        self.punc_engine = None

    def load(self):
        # load vad analyer
        if self._bot_config.vad:
            tag = self._bot_config.vad.tag
            args = self._bot_config.vad.args or {}
            self.vad_analyzer = VADAnalyzerEnvInit.initVADAnalyzerEngine(tag, args)
        else:
            self.vad_analyzer = VADAnalyzerEnvInit.initVADAnalyzerEngine()

        # load asr live engine
        if self._bot_config.asr:
            tag = self._bot_config.asr.tag
            args = self._bot_config.asr.args or {}
            self.asr_engine = ASRLiveEnvInit.getEngine(tag, **args)
        else:
            logging.info("use defualt asr live engine")
            self.asr_engine = ASRLiveEnvInit.getEngine()

        # load punctuation engine
        if self.asr_engine.get_args_dict().get("textnorm", False) is False:
            if self._bot_config.punctuation:
                tag = self._bot_config.punctuation.tag
                args = self._bot_config.punctuation.args or {}
                self.punc_engine = PuncEnvInit.initEngine(tag, **args)

    async def arun(self):
        if self._websocket is None:
            return

        serializer = TranscriptionFrameSerializer()
        self.params = FastapiWebsocketServerParams(
            audio_in_enabled=True,
            audio_out_enabled=False,
            add_wav_header=True,
            vad_enabled=True,
            vad_analyzer=self.vad_analyzer,
            vad_audio_passthrough=True,
            serializer=serializer,
        )
        transport = FastapiWebsocketTransport(
            websocket=self._websocket,
            params=self.params,
        )

        session = Session(**SessionCtx(str(uuid.uuid4())).__dict__)
        asr_live_processor = ASRLiveProcessor(asr=self.asr_engine, session=session)

        processors = [
            transport.input_processor(),
            # FrameLogger(include_frame_types=[AudioRawFrame]),
            # VADAudioSaveProcessor(prefix_name="streaming_asr_vad", pass_raw_audio=True),
            asr_live_processor,
            FrameLogger(include_frame_types=[TextFrame]),
        ]
        if self.punc_engine:
            punc_processor = PunctuationProcessor(engine=self.punc_engine, session=session)
            processors.append(punc_processor)

        processors.append(FrameLogger(include_frame_types=[TextFrame]))
        processors.append(SaveASRText())
        processors.append(transport.output_processor())

        self.task = PipelineTask(
            Pipeline(processors=processors),
            params=PipelineParams(
                allow_interruptions=False,
                enable_metrics=True,
                send_initial_empty_metrics=False,
            ),
        )

        transport.add_event_handler("on_client_connected", self.on_client_connected)
        transport.add_event_handler("on_client_disconnected", self.on_client_disconnected)

        await PipelineRunner(handle_sigint=self._handle_sigint).run(self.task)



# https://docs.comfy.org/tutorials/video/hunyuan/hunyuan-video-1-5
"""
📂 ComfyUI/
├── 📂 models/
│   ├── 📂 text_encoders/
│   │      ├── qwen_2.5_vl_7b_fp8_scaled.safetensors
│   │      └── byt5_small_glyphxl_fp16.safetensors
│   ├── 📂 diffusion_models/
│   │      ├── hunyuanvideo1.5_1080p_sr_distilled_fp16.safetensors
│   │      └── hunyuanvideo1.5_720p_t2v_fp16.safetensors
│   └── 📂 vae/
│          └── hunyuanvideo15_vae_fp16.safetensors
"""

import uuid
import json
import logging
from pathlib import Path
import subprocess

import modal

from . import model_dir, MODEL_VOL, comfyui_out_dir, COMFYUI_OUT_VOL, clear_comfyui_output_dir


# https://github.com/Tencent-Hunyuan/HunyuanVideo-1.5
# https://huggingface.co/tencent/HunyuanVideo-1.5
# https://huggingface.co/Comfy-Org/HunyuanVideo_1.5_repackaged/tree/main/split_files
# modal run src/download_models.py --repo-ids "Comfy-Org/HunyuanVideo_1.5_repackaged" --allow-patterns "*.safetensors"


def link_comfyui_dir():
    """
    ymlink the model to the right ComfyUI directory
    NOTE: need link ckpt file, don't to link models dir
    """
    clear_comfyui_output_dir()

    models_dir = f"{model_dir}/Comfy-Org/HunyuanVideo_1.5_repackaged/split_files"
    for ckpt_file in Path(models_dir).glob("**/*.safetensors"):
        cmd = (
            f"ln -s {ckpt_file} /root/comfy/ComfyUI/models/{ckpt_file.parent.name}/{ckpt_file.name}"
        )
        print(cmd)
        subprocess.run(
            cmd,
            shell=True,
            check=True,
        )


def change_workflow_conf(file_path: Path, **kwargs) -> str:
    workflow_data = json.loads(file_path.read_text())
    client_id = uuid.uuid4().hex

    if kwargs.get("prompt"):
        workflow_data["44"]["inputs"]["text"] = kwargs.get("prompt")

    workflow_data["102"]["inputs"]["filename_prefix"] = client_id
    if kwargs.get("codec"):
        workflow_data["102"]["inputs"]["codec"] = kwargs.get("codec")

    if kwargs.get("width"):
        workflow_data["124"]["inputs"]["width"] = kwargs.get("width")
    if kwargs.get("height"):
        workflow_data["124"]["inputs"]["height"] = kwargs.get("height")
    if kwargs.get("length"):
        workflow_data["124"]["inputs"]["length"] = kwargs.get("length")

    if kwargs.get("steps"):
        workflow_data["128"]["inputs"]["steps"] = kwargs.get("steps")

    print(file_path, workflow_data)
    new_workflow_file = f"{client_id}.json"
    json.dump(workflow_data, Path(new_workflow_file).open("w", encoding="utf-8"))

    return client_id



import kaldi_native_fbank as knf
import numpy as np


class OnlineFbank(knf.OnlineFbank):
    def __init__(self, sample_rate=16000, window_type="povey", dither=0, num_bins=80):
        opts = knf.FbankOptions()
        opts.frame_opts.dither = dither
        opts.frame_opts.samp_freq = sample_rate
        opts.frame_opts.snip_edges = False
        opts.frame_opts.window_type = window_type
        opts.mel_opts.debug_mel = False
        opts.mel_opts.num_bins = num_bins
        super().__init__(opts)

        self.sample_rate = sample_rate
        self.first_available_index = 0

    def accept_waveform(self, samples, is_last=False):
        super().accept_waveform(self.sample_rate, samples)
        if is_last:
            super().input_finished()

    def num_frames_ready(self):
        return super().num_frames_ready - self.first_available_index

    def get_frame(self, index):
        return super().get_frame(index + self.first_available_index)

    def get_frames(self, num_frames=None):
        num_frames = num_frames or self.num_frames_ready()
        frames = []
        for i in range(num_frames):
            frames.append(self.get_frame(i))
        self.first_available_index += num_frames
        return np.stack(frames) if len(frames) > 0 else None

    def get_lfr_frames(self, window_size=7, window_shift=6, neg_mean=0, inv_stddev=1):
        num_lfr_frames = (self.num_frames_ready() - window_size) // window_shift + 1
        if num_lfr_frames <= 0:
            return np.array([])

        num_frames = window_size + (num_lfr_frames - 1) * window_shift
        frames = self.get_frames(num_frames)
        self.first_available_index -= window_size - window_shift
        lfr_frames = np.lib.stride_tricks.as_strided(
            frames,
            shape=(num_lfr_frames, frames.shape[1] * window_size),
            strides=((window_shift * frames.shape[1]) * 4, 4),
        )
        return (lfr_frames + neg_mean) * inv_stddev



import logging
import copy

from dotenv import load_dotenv
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.parallel_pipeline import ParallelPipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger

from src.processors.speech.audio_save_processor import AudioSaveProcessor, SaveAllAudioProcessor
from src.processors.aggregators.user_audio_response import UserAudioResponseAggregator
from src.types.frames.data_frames import (
    TextFrame,
    AudioRawFrame,
    PathAudioRawFrame,
    LLMGenedTokensFrame,
)
from src.common.types import DailyParams
from src.transports.daily import DailyTransport
from src.cmd.bots.base_daily import DailyRoomBot
from src.cmd.bots import register_ai_room_bots
from src.processors.punctuation_processor import PunctuationProcessor
from src.modules.punctuation import PuncEnvInit
from .helper import get_step_audio2_processor, get_step_audio2_llm


load_dotenv(override=True)

"""
TOKENIZERS_PARALLELISM=false python -m src.cmd.bots.main -f config/bots/daily_step_audio2_s2st_bot.json
"""


@register_ai_room_bots.register
class DailyStepAudio2S2STBot(DailyRoomBot):
    """
    daily input/output server bot with step2_audio(translate task)
    """

    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

        self.vad_analyzer = None
        self.audio_llm = None
        self.asr_punc_engine = None

    def load(self):
        self.vad_analyzer = self.get_vad_analyzer()
        llm_conf = self._bot_config.voice_llm or self._bot_config.asr
        if llm_conf:
            self.audio_llm = get_step_audio2_llm(llm_conf)

        # load punctuation engine
        if self._bot_config.punctuation:
            tag = self._bot_config.punctuation.tag
            args = self._bot_config.punctuation.args or {}
            self.asr_punc_engine = PuncEnvInit.initEngine(tag, **args)

    async def arun(self):
        assert self.vad_analyzer is not None

        self.params = DailyParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            vad_enabled=True,
            vad_analyzer=self.vad_analyzer,
            vad_audio_passthrough=True,
        )

        self._voice_processor = None
        if self._bot_config.voice_llm and self.audio_llm:
            # src/processors/voice/step_audio2_processor.py
            self._voice_processor = get_step_audio2_processor(
                self._bot_config.voice_llm,
                session=copy.deepcopy(self.session),
                audio_llm=self.audio_llm,
                processor_class_name="StepS2STProcessor",
            )
            if hasattr(self._voice_processor, "stream_info"):
                stream_info = self._voice_processor.stream_info
                self.params.audio_out_sample_rate = stream_info["sample_rate"]
                self.params.audio_out_channels = stream_info["channels"]
        logging.info(f"params: {self.params}")
        transport = DailyTransport(
            self.args.room_url,
            self.args.token,
            self.args.bot_name,
            self.params,
        )

        asr_processors = []
        if self._bot_config.asr:
            # src/processors/voice/step_audio2_processor.py
            asr_processor = get_step_audio2_processor(
                self._bot_config.asr,
                session=copy.deepcopy(self.session),
                audio_llm=self.audio_llm,
                processor_class_name="StepASRProcessor",
            )

            punc_processor = None
            if self.asr_punc_engine:
                punc_processor = PunctuationProcessor(
                    engine=self.asr_punc_engine, session=self.session
                )

            asr_processors = [
                asr_processor,
                punc_processor,
                FrameLogger(include_frame_types=[TextFrame]),
                transport.output_processor(),
            ]
            asr_processors = [p for p in asr_processors if p is not None]
            logging.info(f"{asr_processors=}")

        stst_processors = []
        if self._voice_processor:
            save_bot_audio_processor = (
                AudioSaveProcessor(prefix_name="bot_speak", pass_raw_audio=True)
                if self._save_audio
                else None
            )
            stst_processors = [
                self._voice_processor,
                FrameLogger(include_frame_types=[TextFrame, AudioRawFrame, LLMGenedTokensFrame]),
                save_bot_audio_processor,
                transport.output_processor(),
            ]
            stst_processors = [p for p in stst_processors if p is not None]
            logging.info(f"{stst_processors=}")

        save_all_records_processor = (
            SaveAllAudioProcessor(
                prefix_name="daily_s2st_bot",
                sample_rate=self.params.audio_in_sample_rate,
                channels=self.params.audio_in_channels,
                sample_width=self.params.audio_in_sample_width,
            )
            if self._save_audio
            else None
        )
        save_speaker_audio_processor = (
            AudioSaveProcessor(prefix_name="user_speak", pass_raw_audio=True)
            if self._save_audio
            else None
        )
        processors = [
            transport.input_processor(),
            save_all_records_processor,
            UserAudioResponseAggregator(),
            FrameLogger(include_frame_types=[AudioRawFrame]),
            save_speaker_audio_processor,
            FrameLogger(include_frame_types=[AudioRawFrame]),
            ParallelPipeline(
                asr_processors,
                stst_processors,
            ),
        ]
        processors = [p for p in processors if p is not None]
        logging.info(f"{processors=}")

        self.task = PipelineTask(
            Pipeline(processors=processors),
            params=PipelineParams(
                allow_interruptions=False,
                enable_metrics=True,
                send_initial_empty_metrics=False,
            ),
        )

        transport.add_event_handlers(
            "on_first_participant_joined",
            [self.on_first_participant_joined, self.on_first_participant_say_hi],
        )
        transport.add_event_handler("on_participant_left", self.on_participant_left)
        transport.add_event_handler("on_call_state_updated", self.on_call_state_updated)

        self.runner = PipelineRunner(handle_sigint=self._handle_sigint)
        await self.runner.run(self.task)

    async def on_first_participant_say_hi(
        self,
        transport: DailyTransport,
        participant,
    ):
        llm_conf = self._bot_config.voice_llm or self._bot_config.llm
        if llm_conf and llm_conf.init_prompt and self._voice_processor:
            await self._voice_processor.say(
                llm_conf.init_prompt,
                temperature=0.1,
                max_new_tokens=1024,
                top_k=20,
                top_p=0.95,
                repetition_penalty=1.1,
            )
